// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading;

using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  public class ImapFetchMessageBodyStream : Stream {
    public static readonly int DefaultFetchBlockSize = 10 * 1024;

    public override bool CanSeek {
      get { return /*!IsClosed &&*/ false; }
    }

    public override bool CanRead {
      get { return !IsClosed /*&& true*/; }
    }

    public override bool CanWrite {
      get { return /*!IsClosed &&*/ false; }
    }

    public override bool CanTimeout {
      get { return true; }
    }

    private bool IsClosed {
      get { return session == null; }
    }

    public override long Position {
      get { RejectDisposed(); return position; }
      set { throw ExceptionUtils.CreateNotSupportedSeekingStream(); }
    }

    public override long Length {
      get { RejectDisposed(); return length; }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return session.ReceiveTimeout; }
      set { RejectDisposed(); session.ReceiveTimeout = value; }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return session.SendTimeout; }
      set { RejectDisposed(); session.SendTimeout = value; }
    }

    protected ImapSession Session {
      get { return session; }
    }

    protected CancellationToken CancellationToken {
      get { return cancellationToken; }
    }

    internal protected long FetchBlockSize {
      get { return fetchBlockSize; }
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  int fetchBlockSize)
    {
      if (session == null)
        throw new ArgumentNullException(nameof(session));
      if (fetchMessageUid == null)
        throw new ArgumentNullException(nameof(fetchMessageUid));
      if (!fetchMessageUid.IsUidSet)
        throw new ArgumentException("must be uid", nameof(fetchMessageUid));
      if (!fetchMessageUid.IsSingle)
        throw new ArgumentException("must be single number", nameof(fetchMessageUid));
      if (fetchBlockSize <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive(nameof(fetchBlockSize), fetchBlockSize);

      this.session = session;
      this.fetchEntireMessage = true;
      this.fetchPeek = fetchPeek;
      this.fetchMessageUid = fetchMessageUid;
      this.fetchBlockSize = (long)fetchBlockSize;
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  string fetchSectionOrHeaderFields,
                                                  ImapPartialRange? fetchRange,
                                                  int fetchBlockSize)
      : this(session,
             fetchPeek,
             fetchMessageUid,
             fetchSectionOrHeaderFields,
             fetchRange,
             fetchBlockSize,
             CancellationToken.None)
    {
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  string fetchSectionOrHeaderFields,
                                                  ImapPartialRange? fetchRange,
                                                  int fetchBlockSize,
                                                  CancellationToken cancellationToken)
      : this(session, fetchPeek, fetchMessageUid, fetchBlockSize)
    {
      this.fetchSectionOrHeaderFields = fetchSectionOrHeaderFields;
      this.fetchRange = fetchRange;

      this.fetchEntireMessage = (string.IsNullOrEmpty(fetchSectionOrHeaderFields) && fetchRange == null);

      if (fetchRange.HasValue && fetchRange.Value.IsLengthSpecified)
        this.length = fetchRange.Value.Length.Value;
      else
        this.length = 0L;

      this.cancellationToken = cancellationToken;
    }

#if NETFRAMEWORK || NETSTANDARD2_0
    public override void Close()
    {
      Dispose(true);
    }
#endif

    protected void DetachFromSession()
    {
      session = null;
    }

    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);

      buffer = null;
      session = null;
    }

    public override void SetLength(long value)
    {
      throw ExceptionUtils.CreateNotSupportedSettingStreamLength();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw ExceptionUtils.CreateNotSupportedSeekingStream();
    }

    public override void Flush()
    {
      // do nothing
    }

    public override void Write(byte[] buffer, int offset, int count)
    {
      throw ExceptionUtils.CreateNotSupportedWritingStream();
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

      cancellationToken.ThrowIfCancellationRequested();

      if (dest == null)
        throw new ArgumentNullException(nameof(dest));
      if (offset < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(offset), offset);
      if (count < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive(nameof(count), count);
      if (dest.Length - count < offset)
        throw ExceptionUtils.CreateArgumentAttemptToAccessBeyondEndOfArray(nameof(offset), dest, offset, count);

      var read = 0;

      for (;;) {
        if (bufRemain == 0) {
          if (fetchedFinalBlock)
            return read;
          else
            FillBuffer();
        }

        if (count <= bufRemain) {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, count);

          bufOffset += count;
          bufRemain -= count;
          position  += count;

          read += count;

          return read;
        }
        else {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, bufRemain);

          offset    += bufRemain;
          count     -= bufRemain;
          position  += bufRemain;

          read += bufRemain;

          bufRemain = 0;
        }
      }
    }

    public void ReadToEnd(Stream destination)
    {
      RejectDisposed();

      if (destination == null)
        throw new ArgumentNullException(nameof(destination));
      if (!destination.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream(nameof(destination));

      // copy buffered data
      if (0 < bufRemain) {
        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }

      // fetch and copy
      for (;;) {
        if (fetchedFinalBlock) {
          bufRemain = 0;
          return;
        }

        FillBuffer();

        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }
    }

    public void ReadToEnd(BinaryWriter destination)
    {
      RejectDisposed();

      if (destination == null)
        throw new ArgumentNullException(nameof(destination));

      // copy buffered data
      if (0 < bufRemain) {
        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }

      // fetch and copy
      for (;;) {
        if (fetchedFinalBlock) {
          bufRemain = 0;
          return;
        }

        FillBuffer();

        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }
    }

    [CLSCompliant(false)]
    internal protected virtual ImapCommandResult Prepare(ImapFetchDataItem fetchDataItem,
                                                         out IImapMessageAttribute messageAttr)
    {
      if (fetchEntireMessage) {
        if (fetchDataItem == null)
          fetchDataItem = ImapFetchDataItem.Rfc822Size;
        else if (!fetchDataItem.ContainsOneOf(ImapFetchDataItem.Rfc822Size))
          fetchDataItem = ImapFetchDataItem.Rfc822Size + fetchDataItem;
      }

      var result = FetchAndFillBuffer(fetchDataItem, out ImapMessage fetched);

      if (result.Succeeded) {
        messageAttr = fetched;

        if (fetchEntireMessage) {
          length = fetched.Rfc822Size;

          fetchedFinalBlock |= (position + bufRemain == length);

          ReportFetchedOctetCount(position + bufRemain, length);
        }
      }
      else {
        messageAttr = null;
      }

      return result;
    }

    private void FillBuffer()
    {
      var result = FetchAndFillBuffer<ImapMessageBody>(null, out _);

      if (result == null)
        return; // nothing to fetch
      else if (result.Failed)
        throw GetFetchFailureException(result);

      if (fetchEntireMessage) {
        fetchedFinalBlock |= (position + bufRemain == length);

        ReportFetchedOctetCount(position + bufRemain, length);
      }
    }

    internal protected virtual void ReportFetchedOctetCount(long fetchedOctetCount, long totalOctetCount)
    {
      // do nothing in this class
    }

    private ImapCommandResult FetchAndFillBuffer<TMessageBody>(ImapFetchDataItem extraFetchDataItem,
                                                               out TMessageBody fetched)
      where TMessageBody : ImapMessageBody
    {
      fetched = null;

      cancellationToken.ThrowIfCancellationRequested();

      ImapCommandResult result;
      TMessageBody[] fetchedMessages;

      try {
        var fetchDataItem = GetNextRangeFetchDataItem(extraFetchDataItem);

        if (fetchDataItem == null) {
          bufOffset = 0;
          bufRemain = 0;
          fetchedFinalBlock = true;

          return null;
        }

        result = session.Fetch(fetchMessageUid,
                               fetchDataItem,
                               out fetchedMessages);
      }
      catch (TimeoutException ex) {
        throw GetTimeoutException(ex);
      }
      catch (ImapException ex) {
        throw GetUnexpectedException(ex);
      }

      if (result.Succeeded) {
        if (fetchedMessages.Length <= 0)
          throw GetNoSuchMessageException(result);

        fetched = fetchedMessages[0];

        var len = fetched.GetFirstBody(ref buffer);

        bufOffset = 0;
        bufRemain = len;

        fetchedFinalBlock = (len < fetchBlockSize);
      }
      else {
        fetched = null;
      }

      return result;
    }

    private ImapFetchDataItem GetNextRangeFetchDataItem(ImapFetchDataItem extraFetchDataItem)
    {
      ImapFetchDataItem bodyTextFetchDataItem;

      if (fetchEntireMessage) {
        bodyTextFetchDataItem = ImapFetchDataItem.BodyText(fetchPeek,
                                                           string.Empty,
                                                           position,
                                                           fetchBlockSize);
      }
      else {
        var fetchStart = fetchRange.HasValue ? fetchRange.Value.Start : 0L;
        var fetchLength = (fetchRange.HasValue && fetchRange.Value.IsLengthSpecified)
          ? Math.Min(fetchRange.Value.Length.Value - position, fetchBlockSize)
          : fetchBlockSize;

        if (fetchLength == 0L)
          return null; // nothing to fetch

        bodyTextFetchDataItem = ImapFetchDataItem.BodyText(fetchPeek,
                                                           fetchSectionOrHeaderFields,
                                                           position + fetchStart,
                                                           fetchLength);
      }

      if (extraFetchDataItem == null)
        return bodyTextFetchDataItem;
      else
        return extraFetchDataItem + bodyTextFetchDataItem;
    }

    protected virtual Exception GetNoSuchMessageException(ImapCommandResult result)
    {
      return new ImapProtocolViolationException("no such message or expunged");
    }

    protected virtual Exception GetFetchFailureException(ImapCommandResult result)
    {
      return new ImapProtocolViolationException(result.ResultText);
    }

    protected virtual Exception GetTimeoutException(TimeoutException ex)
    {
      return ex;
    }

    protected virtual Exception GetUnexpectedException(ImapException ex)
    {
      return ex;
    }

    private void RejectDisposed()
    {
      if (IsClosed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private ImapSession session;
    private readonly bool fetchEntireMessage;
    private readonly bool fetchPeek;
    private readonly ImapSequenceSet fetchMessageUid;
    private readonly string fetchSectionOrHeaderFields;
    private readonly ImapPartialRange? fetchRange;
    private readonly long fetchBlockSize;
    private long length = -1L;
    private long position = 0L;
    private byte[] buffer;
    private int bufOffset = 0;
    private int bufRemain = 0;
    private bool fetchedFinalBlock = false;
    private readonly CancellationToken cancellationToken;
  }
}
