// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client.Session {
  partial class ImapSession {
    /*
     * transaction methods : any state
     */

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Capability()
    {
      return Capability(out _);
    }

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Capability(out ImapCapabilitySet capabilities)
    {
      RejectNonConnectedState();

      capabilities = null;

      using (var t = new CapabilityTransaction(connection)) {
        if (ProcessTransaction(t).Succeeded) {
          SetServerCapabilities(t.Result.Value);
          capabilities = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult NoOp()
    {
      RejectNonConnectedState();

      using (var t = new NoOpTransaction(connection)) {
        return ProcessTransaction(t);
      }
    }

    /// <summary>sends LOGOUT command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Logout()
    {
      CheckDisposed();

      if (state == ImapSessionState.NotConnected)
        return new ImapCommandResult(ImapCommandResultCode.RequestDone,
                                     "already logged out or disconnected");

      using (var t = new LogoutTransaction(connection)) {
        if (ProcessTransaction(t).Code == ImapCommandResultCode.Ok) {
          TransitStateTo(ImapSessionState.NotAuthenticated);
          CloseConnection();
        }

        return t.Result;
      }
    }

    /// <summary>sends ID command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult ID(IReadOnlyDictionary<string, string> clientParameterList)
    {
      return ID(clientParameterList, out _);
    }

    /// <summary>sends ID command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult ID(IReadOnlyDictionary<string, string> clientParameterList, out IReadOnlyDictionary<string, string> serverParameterList)
    {
      RejectNonConnectedState();

      serverParameterList = null;

      using (var t = new IdTransaction(connection)) {
        // client parameter list
        if (clientParameterList != null && 0 < clientParameterList.Count) {
          // RFC 2971 IMAP4 ID extension
          // http://tools.ietf.org/html/rfc2971
          // 3.3. Defined Field Values
          //    Field strings MUST NOT be longer than 30 octets.  Value strings MUST
          //    NOT be longer than 1024 octets.  Implementations MUST NOT send more
          //    than 30 field-value pairs.
          if (30 < clientParameterList.Count)
            throw new ArgumentException("number of client parameter list pairs must be less than or equals to 30", nameof(clientParameterList));

          var idParamsList = new ImapString[clientParameterList.Count * 2];

          try {
            var index = 0;

            foreach (var param in clientParameterList) {
              if (30 < OctetEncoding.SevenBits.GetByteCount(param.Key))
                throw new ArgumentException("field string of client parameter list must be shorter than or equals to 30 octets", nameof(clientParameterList));

              idParamsList[index] = new ImapQuotedString(param.Key);

              if (param.Value == null) {
                idParamsList[index + 1] = ImapNilString.Nil;
              }
              else {
                if (1024 < OctetEncoding.SevenBits.GetByteCount(param.Value))
                  throw new ArgumentException("value string of client parameter list must be shorter than or equals to 1024 octets", nameof(clientParameterList));
                idParamsList[index + 1] = new ImapQuotedString(param.Value);
              }

              index += 2;
            }
          }
          catch (EncoderFallbackException) {
            throw new ArgumentException("both field and value string must not contain non-ASCII chars", nameof(clientParameterList));
          }

          t.RequestArguments["client parameter list"] = new ImapParenthesizedString(idParamsList);
        }

        if (ProcessTransaction(t).Succeeded) {
          this.serverID = t.Result.Value;

          serverParameterList = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPRESS command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Compress(ImapCompressionMechanism compressionMechanism)
    {
      RejectNonConnectedState();

      if (compressionMechanism == null)
        throw new ArgumentNullException(nameof(compressionMechanism));

      using (var t = new CompressTransaction(connection)) {
        t.RequestArguments["compression mechanism"] = compressionMechanism;

        return ProcessTransaction(t);
      }
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Language(out string[] supportedLanguageTags)
    {
      RejectNonConnectedState();

      supportedLanguageTags = null;

      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.2. LANGUAGE Command
       *    If there aren't any arguments, the server SHOULD send an untagged
       *    LANGUAGE response listing the languages it supports.  If the server
       *    is unable to enumerate the list of languages it supports it MAY
       *    return a tagged NO response to the enumeration request.
       */
      using (var t = new LanguageTransaction(connection)) {
        if (ProcessTransaction(t).Succeeded)
          supportedLanguageTags = t.Result.Value.Item1;

        return t.Result;
      }
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Language(CultureInfo cultureInfo, params CultureInfo[] cultureInfos)
    {
      return Language(out _, cultureInfo, cultureInfos);
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Language(out string selectedLanguageTag, CultureInfo cultureInfo, params CultureInfo[] cultureInfos)
    {
      if (cultureInfo == null)
        throw new ArgumentNullException(nameof(cultureInfo));

#if NETFRAMEWORK || NETSTANDARD2_0
      return Language(out selectedLanguageTag, cultureInfo.IetfLanguageTag, cultureInfos.Convert(info => info.IetfLanguageTag));
#else
      return Language(out selectedLanguageTag, cultureInfo.Name, cultureInfos.Convert(info => info.Name));
#endif
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>
    /// valid in any state.
    /// this method sends LANGUAGE command with language of "default".
    /// </remarks>
    public ImapCommandResult Language()
    {
      return Language(out string _);
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>
    /// valid in any state.
    /// this method sends LANGUAGE command with language of "default".
    /// </remarks>
    public ImapCommandResult Language(out string selectedLanguageTag)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 3.2. LANGUAGE Command
       *    The special "default" language range argument indicates a request to
       *    use a language designated as preferred by the server administrator.
       *    The preferred language MAY vary based on the currently active user.
       */
      return Language(out selectedLanguageTag, "default");
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Language(string languageRange, params string[] languageRanges)
    {
      return Language(out _, languageRange, languageRanges);
    }

    /// <summary>sends LANGUAGE command</summary>
    /// <remarks>valid in any state</remarks>
    public ImapCommandResult Language(out string selectedLanguageTag, string languageRange, params string[] languageRanges)
    {
      RejectNonConnectedState();

      if (languageRange == null)
        throw new ArgumentNullException(nameof(languageRange));
      if (languageRange.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(languageRange));

      var languageRangeArguments = new string[1 + (languageRanges == null ? 0 : languageRanges.Length)];

      languageRangeArguments[0] = languageRange;

      if (languageRanges != null) {
        for (var i = 0; i < languageRanges.Length; i++) {
          if (languageRanges[i] == null)
            throw new ArgumentNullException(string.Format("{0}[{1}]", nameof(languageRanges), i));
          if (languageRanges[i].Length == 0)
            throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(string.Format("{0}[{1}]", nameof(languageRanges), i));

          languageRangeArguments[1 + i] = languageRanges[i];
        }
      }

      selectedLanguageTag = null;

      using (var t = new LanguageTransaction(connection)) {
        t.RequestArguments["language range arguments"] = new ImapStringList(languageRangeArguments);

        if (ProcessTransaction(t).Succeeded) {
          this.selectedLanguage = t.Result.Value.Item1[0];

          if (t.Result.Value.Item2 != null)
            this.namespaces = t.Result.Value.Item2;

          selectedLanguageTag = t.Result.Value.Item1[0];
        }

        return t.Result;
      }
    }
  }
}
