// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  partial class ImapSession {
    /*
     * transaction methods : authenticated/selected state
     */

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, string resourceName, long resourceLimit)
    {
      return SetQuota(quotaRoot, new[] {new ImapQuotaResource(resourceName, resourceLimit)}, out _);
    }

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, string resourceName, long resourceLimit, out ImapQuota changedQuota)
    {
      return SetQuota(quotaRoot, new[] {new ImapQuotaResource(resourceName, resourceLimit)}, out changedQuota);
    }

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, ImapQuotaResource resourceLimit)
    {
      return SetQuota(quotaRoot, new[] {resourceLimit}, out _);
    }

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, ImapQuotaResource resourceLimit, out ImapQuota changedQuota)
    {
      return SetQuota(quotaRoot, new[] {resourceLimit}, out changedQuota);
    }

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, ImapQuotaResource[] resourceLimits)
    {
      return SetQuota(quotaRoot, resourceLimits, out _);
    }

    /// <summary>sends SETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult SetQuota(string quotaRoot, ImapQuotaResource[] resourceLimits, out ImapQuota changedQuota)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));
      if (resourceLimits == null)
        throw new ArgumentNullException(nameof(resourceLimits));

      var listOfResourceLimits = new ImapString[resourceLimits.Length * 2];
      var index = 0;

      foreach (var resourceLimit in resourceLimits) {
        if (resourceLimit == null)
          throw new ArgumentException("contains null", nameof(resourceLimits));

        listOfResourceLimits[index++] = resourceLimit.Name;
        listOfResourceLimits[index++] = resourceLimit.Limit.ToString();
      }

      RejectNonAuthenticatedState();

      changedQuota = null;

      using (var t = new SetQuotaTransaction(connection)) {
        // quota root
        t.RequestArguments["quota root"] = new ImapQuotedString(quotaRoot);
        // list of resource limits
        t.RequestArguments["list of resource limits"] = new ImapParenthesizedString(listOfResourceLimits);

        if (ProcessTransaction(t).Succeeded)
          changedQuota = t.Result.Value;

        return t.Result;
      }
    }

    /// <summary>sends GETQUOTA command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult GetQuota(string quotaRoot, out ImapQuota quota)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException(nameof(quotaRoot));

      RejectNonAuthenticatedState();

      quota = null;

      using (var t = new GetQuotaTransaction(connection)) {
        // quota root
        t.RequestArguments["quota root"] = new ImapQuotedString(quotaRoot);

        if (ProcessTransaction(t).Succeeded)
          quota = t.Result.Value;

        return t.Result;
      }
    }

    /// <summary>sends GETQUOTAROOT command</summary>
    /// <remarks>valid in authenticated/selected state (undocumented)</remarks>
    public ImapCommandResult GetQuotaRoot(ImapMailbox mailbox, out IReadOnlyDictionary<string, ImapQuota[]> quotaRoots)
    {
      ValidateMailboxRelationship(mailbox);

      return GetQuotaRoot(mailbox.Name, out quotaRoots);
    }

    public ImapCommandResult GetQuotaRoot(string mailboxName, out IReadOnlyDictionary<string, ImapQuota[]> quotaRoots)
    {
      RejectNonAuthenticatedState();

      var argMailboxName = ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName);

      quotaRoots = null;

      using (var t = new GetQuotaRootTransaction(connection)) {
        // mailbox name
        t.RequestArguments["mailbox name"] = argMailboxName;

        if (ProcessTransaction(t).Succeeded)
          quotaRoots = t.Result.Value;

        return t.Result;
      }
    }

    public ImapCommandResult Idle(TimeSpan timeout)
    {
      return IdleInternal((int)timeout.TotalMilliseconds, null, null);
    }

    public ImapCommandResult Idle(int millisecondsTimeout)
    {
      return IdleInternal(millisecondsTimeout, null, null);
    }

    public ImapCommandResult Idle(object state, ImapKeepIdleCallback keepIdleCallback)
    {
      return Idle(Timeout.Infinite, state, keepIdleCallback);
    }

    public ImapCommandResult Idle(TimeSpan idleTimeout, object state, ImapKeepIdleCallback keepIdleCallback)
    {
      return Idle((int)idleTimeout.TotalMilliseconds, state, keepIdleCallback);
    }

    public ImapCommandResult Idle(int idleMillisecondsTimeout, object state, ImapKeepIdleCallback keepIdleCallback)
    {
      if (keepIdleCallback == null)
        throw new ArgumentNullException(nameof(keepIdleCallback));

      return IdleInternal(idleMillisecondsTimeout, state, keepIdleCallback);
    }

    private ImapCommandResult IdleInternal(int idleMillisecondsTimeout, object keepIdleState, ImapKeepIdleCallback keepIdleCallback)
    {
      if (idleMillisecondsTimeout < -1)
        throw ExceptionUtils.CreateArgumentMustBeGreaterThanOrEqualTo(-1, nameof(idleMillisecondsTimeout), idleMillisecondsTimeout);

      RejectNonAuthenticatedState();
      RejectIdling();

      using (var cts = new CancellationTokenSource(idleMillisecondsTimeout)) {
        using (var t = new IdleTransaction(connection, keepIdleState, keepIdleCallback, cts.Token)) {
          PreProcessTransaction(t, handlesIncapableAsException);

          PostProcessTransaction(ProcessTransactionInternal(t));

          return t.Result;
        }
      }
    }

    public Task<ImapCommandResult> IdleAsync(object keepIdleState, ImapKeepIdleCallback keepIdleCallback)
    {
      return IdleAsync(keepIdleState, keepIdleCallback, CancellationToken.None);
    }

    public Task<ImapCommandResult> IdleAsync(CancellationToken stopIdleCancellationToken)
    {
      return IdleAsync(null, null, stopIdleCancellationToken);
    }

    public Task<ImapCommandResult> IdleAsync(object keepIdleState, ImapKeepIdleCallback keepIdleCallback, CancellationToken stopIdleCancellationToken)
    {
      RejectNonAuthenticatedState();
      RejectIdling();

      var idleTransaction = new IdleTransaction(connection, keepIdleState, keepIdleCallback, stopIdleCancellationToken);

      try {
        PreProcessTransaction(idleTransaction, handlesIncapableAsException);
      }
      catch {
        idleTransaction.Dispose();

        throw;
      }

      var idleTask = Task.Run(() => PostProcessTransaction(ProcessTransactionInternal(idleTransaction)));

      idleTask.ContinueWith(task => idleTransaction.Dispose());

      // wait for IDLE started or respond tagged error response
      idleTransaction.WaitForIdleStateChanged();

      return idleTask;
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Namespace()
    {
      return Namespace(out _);
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Namespace(out ImapNamespace namespaces)
    {
      RejectNonAuthenticatedState();

      namespaces = null;

      using (var t = new NamespaceTransaction(connection)) {
        if (ProcessTransaction(t).Succeeded) {
          this.namespaces = t.Result.Value;
          namespaces = t.Result.Value;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator()
    {
      return Comparator(out _);
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(out ImapCollationAlgorithm activeComparator)
    {
      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 4.7. COMPARATOR Command
       *    When issued with no arguments, it results in a
       *    COMPARATOR response indicating the currently active comparator.
       */
      activeComparator = null;

      using (var t = new ComparatorTransaction(connection)) {
        if (ProcessTransaction(t).Succeeded) {
          this.activeComparator = t.Result.Value.Item1;

          activeComparator = t.Result.Value.Item1;
        }

        return t.Result;
      }
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      return Comparator(out _, out _, comparator, comparators);
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(out ImapCollationAlgorithm[] matchingComparators, ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      return Comparator(out _, out matchingComparators, comparator, comparators);
    }

    /// <summary>sends COMPARATOR command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public ImapCommandResult Comparator(out ImapCollationAlgorithm activeComparator, out ImapCollationAlgorithm[] matchingComparators, ImapCollationAlgorithm comparator, params ImapCollationAlgorithm[] comparators)
    {
      RejectNonAuthenticatedState();

      if (comparator == null)
        throw new ArgumentNullException(nameof(comparator));

      var comparatorArguments = new ImapCollationAlgorithm[1 + (comparators == null ? 0 : comparators.Length)];

      comparatorArguments[0] = comparator;

      if (comparators != null) {
        for (var i = 0; i < comparators.Length; i++) {
          if (comparators[i] == null)
            throw new ArgumentNullException(string.Format("{0}[{1}]", nameof(comparators), i));

          comparatorArguments[1 + i] = comparators[i];
        }
      }

      activeComparator = null;
      matchingComparators = null;

      /*
       * RFC 5255 - Internet Message Access Protocol Internationalization
       * http://tools.ietf.org/html/rfc5255
       * 4.7. COMPARATOR Command
       *    When issued with one or more comparator arguments, it changes the
       *    active comparator as directed.
       */
      using (var t = new ComparatorTransaction(connection)) {
        t.RequestArguments["comparator order arguments"] = new ImapStringList(comparatorArguments.Convert(comp => new ImapQuotedString(comp.ToString())));

        if (ProcessTransaction(t).Succeeded) {
          this.activeComparator = t.Result.Value.Item1;

          activeComparator = t.Result.Value.Item1;
          matchingComparators = t.Result.Value.Item2 ??
#if NET45
            ArrayExtensions.Empty<ImapCollationAlgorithm>();
#else
            Array.Empty<ImapCollationAlgorithm>();
#endif
        }

        return t.Result;
      }
    }

#region "GETMETADATA"
#region "server entries"
    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string entrySpecifier, out ImapMetadata[] metadata)
    {
      return GetMetadataInternal(new[] {entrySpecifier}, null, out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string[] entrySpecifiers, out ImapMetadata[] metadata)
    {
      return GetMetadataInternal(entrySpecifiers, null, out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string entrySpecifier, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      if (options == null)
        throw new ArgumentNullException(nameof(options));

      return GetMetadataInternal(new[] {entrySpecifier}, options, out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string[] entrySpecifiers, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      if (options == null)
        throw new ArgumentNullException(nameof(options));

      return GetMetadataInternal(entrySpecifiers, options, out metadata);
    }

    private ImapCommandResult GetMetadataInternal(string[] entrySpecifiers, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      return GetMetadataInternal(null,
                                 entrySpecifiers,
                                 options,
                                 out metadata);
    }
#endregion

#region "mailbox entries"
    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailbox mailbox, string entrySpecifier, out ImapMetadata[] metadata)
    {
      ValidateMailboxRelationship(mailbox);

      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailbox.Name),
                                 new[] {entrySpecifier},
                                 null,
                                 out metadata);
    }

    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailbox mailbox, string[] entrySpecifiers, out ImapMetadata[] metadata)
    {
      ValidateMailboxRelationship(mailbox);

      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailbox.Name),
                                 entrySpecifiers,
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName, string entrySpecifier, out ImapMetadata[] metadata)
    {
      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName),
                                 new[] {entrySpecifier},
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName, string[] entrySpecifiers, out ImapMetadata[] metadata)
    {
      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName),
                                 entrySpecifiers,
                                 null,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailbox mailbox, string entrySpecifier, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      if (options == null)
        throw new ArgumentNullException(nameof(options));

      ValidateMailboxRelationship(mailbox);

      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailbox.Name),
                                 new[] {entrySpecifier},
                                 options,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(ImapMailbox mailbox, string[] entrySpecifiers, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      if (options == null)
        throw new ArgumentNullException(nameof(options));

      ValidateMailboxRelationship(mailbox);

      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailbox.Name),
                                 entrySpecifiers,
                                 options,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName, string entrySpecifier, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      if (options == null)
        throw new ArgumentNullException(nameof(options));

      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName),
                                 new[] {entrySpecifier},
                                 options,
                                 out metadata);
    }

    /// <summary>sends GETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult GetMetadata(string mailboxName, string[] entrySpecifiers, ImapGetMetadataOptions options, out ImapMetadata[] metadata)
    {
      if (options == null)
        throw new ArgumentNullException(nameof(options));

      return GetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName),
                                 entrySpecifiers,
                                 options,
                                 out metadata);
    }
#endregion

    private ImapCommandResult GetMetadataInternal(ImapString mailboxName,
                                                  string[] entrySpecifiers,
                                                  ImapGetMetadataOptions options,
                                                  out ImapMetadata[] metadata)
    {
      if (entrySpecifiers == null)
        throw new ArgumentNullException(nameof(entrySpecifiers));
      if (entrySpecifiers.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyArray(nameof(entrySpecifiers));

      if (handlesIncapableAsException)
        CheckServerCapabilityMetadata(mailboxName != null);

      metadata = null;

      using (var t = new GetMetadataTransaction(connection)) {
        // mailbox-name
        if (mailboxName == null)
          t.RequestArguments["mailbox-name"] = ImapQuotedString.Empty;
        else
          t.RequestArguments["mailbox-name"] = mailboxName;

        // options
        if (options != null)
          t.RequestArguments["options"] = options;

        // entry-specifier
        if (entrySpecifiers.Length == 1)
          t.RequestArguments["entry-specifier"] = entrySpecifiers[0];
        else
          t.RequestArguments["entry-specifier"] = new ImapParenthesizedString(entrySpecifiers);

        if (ProcessTransaction(t).Succeeded)
          metadata = t.Result.Value;

        return t.Result;
      }
    }
#endregion

#region "SETMETADATA"
#region "server entries"
    /// <summary>sends SETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(params ImapMetadata[] metadata)
    {
      return SetMetadataInternal(null,
                                 metadata);
    }

    /// <summary>sends SETMETADATA command and sets NIL value to remove entries</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA or METADATA-SERVER extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(string[] entrySpecifiers)
    {
      return SetMetadataInternal(null,
                                 entrySpecifiers.Convert(ImapMetadata.CreateNil));
    }
#endregion

#region "mailbox entries"
    /// <summary>sends SETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(ImapMailbox mailbox, params ImapMetadata[] metadata)
    {
      ValidateMailboxRelationship(mailbox);

      return SetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailbox.Name),
                                 metadata);
    }

    /// <summary>sends SETMETADATA command</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(string mailboxName, params ImapMetadata[] metadata)
    {
      return SetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName),
                                 metadata);
    }

    /// <summary>sends SETMETADATA command and sets NIL value to remove entries</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(ImapMailbox mailbox, params string[] entrySpecifiers)
    {
      ValidateMailboxRelationship(mailbox);

      return SetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailbox.Name),
                                 entrySpecifiers.Convert(ImapMetadata.CreateNil));
    }

    /// <summary>sends SETMETADATA command and sets NIL value to remove entries</summary>
    /// <remarks>
    /// valid in authenticated/selected state.
    /// this method will fail if server does not support METADATA extension.
    /// </remarks>
    public ImapCommandResult SetMetadata(string mailboxName, params string[] entrySpecifiers)
    {
      return SetMetadataInternal(ImapMailboxNameString.CreateMailboxNameNonEmpty(mailboxName),
                                 entrySpecifiers.Convert(ImapMetadata.CreateNil));
    }
#endregion

    private ImapCommandResult SetMetadataInternal(ImapMailboxNameString mailboxName,
                                                  ImapMetadata[] metadata)
    {
      if (metadata == null)
        throw new ArgumentNullException(nameof(metadata));

      if (handlesIncapableAsException)
        CheckServerCapabilityMetadata(mailboxName != null);

      using (var t = new SetMetadataTransaction(connection)) {
        // mailbox-name
        if (mailboxName == null)
          t.RequestArguments["mailbox-name"] = ImapQuotedString.Empty;
        else
          t.RequestArguments["mailbox-name"] = mailboxName;

        // list of entry, values
        var listOfEntryAndValues = new ImapString[metadata.Length * 2];

        for (var i = 0; i < metadata.Length; i++) {
          var m = metadata[i];

          if (m == null)
            throw new ArgumentException("contains null", nameof(metadata));

          listOfEntryAndValues[i * 2] = m.EntryName;

          if (m.Value == null)
            listOfEntryAndValues[i * 2 + 1] = ImapNilString.Nil;
          else
            listOfEntryAndValues[i * 2 + 1] = m.Value;
        }

        t.RequestArguments["list of entry, values"] = new ImapParenthesizedString(listOfEntryAndValues);

        return ProcessTransaction(t);
      }
    }
#endregion

    private void CheckServerCapabilityMetadata(bool requireMailboxAnnotations)
    {
      /*
       * RFC 5464 - The IMAP METADATA Extension
       * http://tools.ietf.org/html/rfc5464
       * 1. Introduction and Overview
       *
       *    A server that supports both server and mailbox annotations indicates
       *    the presence of this extension by returning "METADATA" as one of the
       *    supported capabilities in the CAPABILITY command response.
       * 
       *    A server that supports only server annotations indicates the presence
       *    of this extension by returning "METADATA-SERVER" as one of the
       *    supported capabilities in the CAPABILITY command response.
       */
      if (serverCapabilities.Contains(ImapCapability.Metadata))
        return;
      else if (requireMailboxAnnotations)
        throw new ImapIncapableException(ImapCapability.Metadata);
      else if (!serverCapabilities.Contains(ImapCapability.MetadataServer))
        throw new ImapIncapableException(ImapCapability.MetadataServer);
    }
  }
}
