// 
// Copyright (c) 2008 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public abstract class ImapMessageInfoBase {
    public ImapClient Client {
      get { return client; }
    }

    public ImapOpenedMailboxInfo Mailbox {
      get { CheckAttachedToSelectedMailbox(); return mailbox; }
    }

    public bool IsAvailable {
      get { return mailbox != null && client.OpeningMailbox == mailbox; }
    }

    public long UidValidity {
      get; protected set;
    }

    private readonly ImapClient client;
    private ImapOpenedMailboxInfo mailbox;

    protected ImapMessageInfoBase(ImapOpenedMailboxInfo mailbox)
    {
      if (mailbox == null)
        throw new ArgumentNullException(nameof(mailbox));

      this.client = mailbox.Client;
      this.mailbox = mailbox;
      this.UidValidity = mailbox.UidValidity; // save current UIDVALIDITY
    }

    protected abstract ImapSequenceSet GetSequenceOrUidSet();

    protected void CheckAttachedToSelectedMailbox()
    {
      if (!IsAvailable)
        throw new ImapUnavailableException();
    }

    protected void CheckMessageNumberAvailable(ImapSequenceSet sequenceOrUidSet)
    {
      CheckAttachedToSelectedMailbox();

      mailbox.CheckUidValidity(mailbox.UidValidity, sequenceOrUidSet);
    }

    internal void DetachFromMailbox()
    {
#if DEBUG
      client.ThrowIfNotInExclusiveOperation();
#endif

      mailbox = null;
    }

    /*
     * operations
     */
    public void AddFlags(ImapMessageFlag flag,
                         params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.AddFlags(flag, flags));
    }

    public void AddKeywords(string keyword,
                            params string[] keywords)
    {
      Store(ImapStoreDataItem.AddFlags(keyword, keywords));
    }

    public void AddFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.AddFlags(flagsAndKeywords));
    }

    public Task AddFlagsAsync(ImapMessageFlag flag,
                              params ImapMessageFlag[] flags)
    {
      return StoreAsync(ImapStoreDataItem.AddFlags(flag, flags));
    }

    public Task AddKeywordsAsync(string keyword,
                                 params string[] keywords)
    {
      return StoreAsync(ImapStoreDataItem.AddFlags(keyword, keywords));
    }

    public Task AddFlagsAsync(IImapMessageFlagSet flagsAndKeywords)
    {
      return StoreAsync(ImapStoreDataItem.AddFlags(flagsAndKeywords));
    }

    public void RemoveFlags(ImapMessageFlag flag,
                            params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.RemoveFlags(flag, flags));
    }

    public void RemoveKeywords(string keyword,
                               params string[] keywords)
    {
      Store(ImapStoreDataItem.RemoveFlags(keyword, keywords));
    }

    public void RemoveFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.RemoveFlags(flagsAndKeywords));
    }

    public Task RemoveFlagsAsync(ImapMessageFlag flag,
                                 params ImapMessageFlag[] flags)
    {
      return StoreAsync(ImapStoreDataItem.RemoveFlags(flag, flags));
    }

    public Task RemoveKeywordsAsync(string keyword,
                                    params string[] keywords)
    {
      return StoreAsync(ImapStoreDataItem.RemoveFlags(keyword, keywords));
    }

    public Task RemoveFlagsAsync(IImapMessageFlagSet flagsAndKeywords)
    {
      return StoreAsync(ImapStoreDataItem.RemoveFlags(flagsAndKeywords));
    }

    public void SetFlags(ImapMessageFlag flag,
                          params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.ReplaceFlags(flag, flags));
    }

    public void SetKeywords(string keyword,
                            params string[] keywords)
    {
      Store(ImapStoreDataItem.ReplaceFlags(keyword, keywords));
    }

    public void SetFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.ReplaceFlags(flagsAndKeywords));
    }

    public Task SetFlagsAsync(ImapMessageFlag flag,
                              params ImapMessageFlag[] flags)
    {
      return StoreAsync(ImapStoreDataItem.ReplaceFlags(flag, flags));
    }

    public Task SetKeywordsAsync(string keyword,
                                 params string[] keywords)
    {
      return StoreAsync(ImapStoreDataItem.ReplaceFlags(keyword, keywords));
    }

    public Task SetFlagsAsync(IImapMessageFlagSet flagsAndKeywords)
    {
      return StoreAsync(ImapStoreDataItem.ReplaceFlags(flagsAndKeywords));
    }

    public void Store(ImapStoreDataItem storeDataItem)
    {
      if (storeDataItem == null)
        throw new ArgumentNullException(nameof(storeDataItem));

      client.RunExclusively(() => {
        var sequenceOrUidSet = GetSequenceOrUidSet();

        if (sequenceOrUidSet.IsEmpty)
          return; // do nothing

        StoreCore(sequenceOrUidSet, storeDataItem);
      });
    }

    public Task StoreAsync(ImapStoreDataItem storeDataItem)
    {
      if (storeDataItem == null)
        throw new ArgumentNullException(nameof(storeDataItem));

      return client.StartTaskExclusively(() => {
        var sequenceOrUidSet = GetSequenceOrUidSet();

        if (sequenceOrUidSet.IsEmpty)
          return; // do nothing

        StoreCore(sequenceOrUidSet, storeDataItem);
      }, CancellationToken.None);
    }

    protected void StoreCore(ImapSequenceSet sequenceOrUidSet,
                             ImapStoreDataItem storeDataItem)
    {
      CheckMessageNumberAvailable(sequenceOrUidSet);

      mailbox.ProcessResult(client.Session.Store(sequenceOrUidSet, storeDataItem));
    }

    private static ImapStoreDataItem markAsDeletedStoreDataItem = ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted);

    public void MarkAsDeleted()
    {
      Store(markAsDeletedStoreDataItem);
    }

    public Task MarkAsDeletedAsync()
    {
      return StoreAsync(markAsDeletedStoreDataItem);
    }

    private static ImapStoreDataItem markAsSeenStoreDataItem = ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen);

    public void MarkAsSeen()
    {
      Store(markAsSeenStoreDataItem);
    }

    public Task MarkAsSeenAsync()
    {
      return StoreAsync(markAsSeenStoreDataItem);
    }

    public void Delete()
    {
      client.RunExclusively(DeleteCore);
    }

    public Task DeleteAsync()
    {
      return client.StartTaskExclusively(DeleteCore, CancellationToken.None);
    }

    private void DeleteCore()
    {
      var sequenceOrUidSet = GetSequenceOrUidSet();

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      StoreCore(sequenceOrUidSet, markAsDeletedStoreDataItem);

      if (sequenceOrUidSet.IsUidSet && client.IsCapabilityAvailable(ImapCapability.UidPlus))
        mailbox.ProcessResult(client.Session.UidExpunge(sequenceOrUidSet));
      else
        mailbox.ProcessResult(client.Session.Expunge());

      if (IsAvailable)
        mailbox.Detach(this);
    }

    public void CopyTo(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy, destinationMailbox));
    }

    public Task CopyToAsync(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      return client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy, destinationMailbox),
                                         CancellationToken.None);
    }

    public void MoveTo(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailbox.Mailbox))
        return; // do nothing

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move, destinationMailbox));
    }

    public Task MoveToAsync(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException(nameof(destinationMailbox));

      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailbox.Mailbox)) {
        // do nothing
        var tcs = new TaskCompletionSource<object>();

        tcs.SetResult(null);

        return tcs.Task;
      }

      destinationMailbox.ThrowIfNonExistentOrUnselectable();

      return client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move, destinationMailbox),
                                         CancellationToken.None);
    }

    private void CopyToMoveToCore(ImapSession.CopyMoveCommand command,
                                  ImapMailboxInfo destinationMailbox)
    {
      var sequenceOrUidSet = GetSequenceOrUidSet();

      if (sequenceOrUidSet.IsEmpty)
        return; // do nothing

      CheckMessageNumberAvailable(sequenceOrUidSet);

      var move = command == ImapSession.CopyMoveCommand.Move;
      var markAsDelete = false;

      if (destinationMailbox.Client == this.client) {
        // same session
        if (move && !client.IsCapabilityAvailable(ImapCapability.Move)) {
          // use COPY+STORE instead of MOVE
          command = ImapSession.CopyMoveCommand.Copy;
          markAsDelete = true;
        }

        mailbox.ProcessResult(client.Session.Copy(command, sequenceOrUidSet, destinationMailbox.Mailbox));
      }
      else {
        // different session
        if (move)
          markAsDelete = true;

        var appendMessage = this as ImapMessageInfo;

        if (appendMessage == null) {
          var appendMessages = new ConcreteMessageInfoList(mailbox,
                                                           ImapMessageFetchAttributeOptions.None,
                                                           sequenceOrUidSet);

          foreach (var message in appendMessages.YieldMessages()) { // not to SetBusy
            destinationMailbox.CopyMessageAcrossSession(message, false);
          }
        }
        else {
          destinationMailbox.CopyMessageAcrossSession(appendMessage, false);
        }
      }

      if (markAsDelete)
        StoreCore(sequenceOrUidSet, markAsDeletedStoreDataItem);

      if (move && IsAvailable)
        mailbox.Detach(this);
    }

    public void CopyTo(string destinationMailboxName)
    {
      CopyTo(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo CopyTo(string destinationMailboxName,
                                  bool tryCreate)
    {
      ImapClient.GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName");

      return client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy,
                                                          destinationMailboxName,
                                                          tryCreate));
    }

    public Task CopyToAsync(string destinationMailboxName)
    {
      return CopyToAsync(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public Task<ImapMailboxInfo> CopyToAsync(string destinationMailboxName,
                                             bool tryCreate)
    {
      ImapClient.GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName");

      return client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Copy,
                                                                destinationMailboxName,
                                                                tryCreate),
                                         CancellationToken.None);
    }

    public void MoveTo(string destinationMailboxName)
    {
      MoveTo(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo MoveTo(string destinationMailboxName,
                                  bool tryCreate)
    {
      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailboxName))
        return null; // do nothing

      ImapClient.GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName");

      return client.RunExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move,
                                                          destinationMailboxName,
                                                          tryCreate));
    }

    public Task MoveToAsync(string destinationMailboxName)
    {
      return MoveToAsync(destinationMailboxName, false);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public Task<ImapMailboxInfo> MoveToAsync(string destinationMailboxName,
                                             bool tryCreate)
    {
      CheckAttachedToSelectedMailbox();

      if (mailbox.Mailbox.NameEquals(destinationMailboxName)) {
        // do nothing
        var tcs = new TaskCompletionSource<ImapMailboxInfo>();

        tcs.SetResult(null);

        return tcs.Task;
      }

      ImapClient.GetValidatedMailboxName(destinationMailboxName, "destinationMailboxName");

      return client.StartTaskExclusively(() => CopyToMoveToCore(ImapSession.CopyMoveCommand.Move,
                                                                destinationMailboxName,
                                                                tryCreate),
                                         CancellationToken.None);
    }

    private ImapMailboxInfo CopyToMoveToCore(ImapSession.CopyMoveCommand command,
                                             string destinationMailboxName,
                                             bool tryCreate)
    {
      var sequenceOrUidSet = GetSequenceOrUidSet();

      if (sequenceOrUidSet.IsEmpty)
        return null; // do nothing

      var move = (command == ImapSession.CopyMoveCommand.Move);
      var markAsDelete = false;

      if (move && !client.IsCapabilityAvailable(ImapCapability.Move)) {
        // use COPY+STORE instead of MOVE
        command = ImapSession.CopyMoveCommand.Copy;
        markAsDelete = true;
      }

      CheckMessageNumberAvailable(sequenceOrUidSet);

      ImapMailbox createdMailbox = null;

      if (tryCreate) {
        mailbox.ProcessResult(client.Session.Copy(command,
                                                  sequenceOrUidSet,
                                                  destinationMailboxName,
                                                  out createdMailbox));
      }
      else {
        mailbox.ProcessResult(client.Session.Copy(command,
                                                  sequenceOrUidSet,
                                                  destinationMailboxName),
                              delegate(ImapTaggedStatusResponse status) {
          if (status.ResponseText.Code == ImapResponseCode.TryCreate)
            throw new ImapMailboxNotFoundException(destinationMailboxName);
          else
            return true;
        });
      }

      var createdMailboxInfo = (createdMailbox == null)
        ? null
        : client.CreateFlagsRetrieved(createdMailbox);

      if (markAsDelete)
        StoreCore(sequenceOrUidSet, markAsDeletedStoreDataItem);

      if (move && IsAvailable)
        mailbox.Detach(this);

      return createdMailboxInfo;
    }
  }
}
