using System;
using System.IO;
using System.Text;
using System.Threading;
using NUnit.Framework;

using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionCommandsAppendTests {
    [Test]
    public void TestAppend()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_NonSynchronizedLiterals()
    {
      TestCases.WithMailboxSelectedSession(new[] {"LITERAL+"}, (session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_TryCreate()
    {
      TestCases.WithMailboxSelectedSession((session, server, expectedAuthority) => {
        session.TransactionTimeout = 500;
        server.EnqueueLiteralContinuationResponse("APPEND INBOX.appendto ",
                                                  "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n",
                                                  false);
        server.EnqueueTaggedResponse("CREATE INBOX.appendto\r\n",
                                     "$tag OK CREATE completed\r\n");
        server.EnqueueTaggedResponse("APPEND INBOX.appendto ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Draft));

        ImapMailbox created;

        Assert.IsTrue((bool)session.Append(message, "INBOX.appendto", out created));

        Assert.AreEqual("INBOX.appendto", created.Name);
        Assert.AreEqual(expectedAuthority + "INBOX.appendto", created.Url.AbsoluteUri);
        Assert.IsNotNull(created.Flags);
        Assert.IsNotNull(created.ApplicableFlags);
        Assert.IsNotNull(created.PermanentFlags);

        server.DequeueRequest(); // APPEND - NO [TRYCREATE]
        server.DequeueRequest(); // CREATE

        Assert.AreEqual("0006 APPEND INBOX.appendto (\\Draft $label2) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_TryCreate_NoReferralResponseCode()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueLiteralContinuationResponse("APPEND SHARED/FOO ",
                                                  "$tag NO [REFERRAL IMAP://user;AUTH=*@SERVER2/SHARED/FOO] Remote mailbox. Try SERVER2.\r\n",
                                                  false);

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Draft));

        ImapMailbox created;

        Assert.IsFalse((bool)session.Append(message, "SHARED/FOO", out created));

        Assert.IsNull(created);

        Assert.AreEqual("0004 APPEND SHARED/FOO (\\Draft $label2) {33}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_TryCreate_DontTryCreate()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueLiteralContinuationResponse("APPEND INBOX.appendto ",
                                                  "$tag NO [TRYCREATE] Mailbox doesn't exist: INBOX.appendto\r\n",
                                                  false);

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new DateTimeOffset(2008, 2, 25, 15, 1, 12, new TimeSpan(+9, 0, 0)));

        var internalDate = "\"25-Feb-2008 15:01:12 +0900\"";

        Assert.IsFalse((bool)session.Append(message, "INBOX.appendto"));

        Assert.AreEqual("0004 APPEND INBOX.appendto " + internalDate + " {33}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_RemoveNonApplicableFlags()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen,
                                                                    ImapMessageFlag.Recent,
                                                                    ImapMessageFlag.AllowedCreateKeywords));

        Assert.IsTrue((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppend_WithAppendUidResponseCode()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.Append(message, "INBOX", out appended));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(appended);
        Assert.AreEqual(38505L, appended.UidValidity);
        Assert.AreEqual(3955L, appended.ToNumber());
        Assert.AreEqual("3955", appended.ToString());

        TestUtils.Assert.IsSerializableBinaryFormat(appended, delegate(ImapAppendedUidSet deserialized) {
          Assert.AreEqual(38505L, deserialized.UidValidity);
          Assert.AreEqual("3955", deserialized.ToString());
        });
      });
    }

    [Test]
    public void TestAppend_Continuation_ResponseNo()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        server.EnqueueLiteralContinuationResponse("APPEND INBOX ",
                                                  "$tag NO APPEND failed\r\n",
                                                  false);

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message"),
                                            new ImapMessageFlagSet(ImapMessageFlag.Seen));

        Assert.IsFalse((bool)session.Append(message, "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX (\\Seen) {33}\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppendMultiple()
    {
      AppendMultiple(false);
    }

    [Test]
    public void TestAppendMultiple_WithAppendUidResponseCode()
    {
      AppendMultiple(true);
    }

    private void AppendMultiple(bool respondAppendUid)
    {
      TestCases.WithMailboxSelectedSession(new[] {"MULTIAPPEND"}, (session, server) => {
        server.EnqueueLiteralContinuationResponse("+ Ready for literal data\r\n", true);
        server.EnqueueLiteralContinuationResponse("+ Ready for literal data\r\n", true);

        if (respondAppendUid)
          server.EnqueueTaggedResponse("APPEND saved-messages ",
                                       "$tag OK [APPENDUID 38505 3955:3956] APPEND completed\r\n");
        else
          server.EnqueueTaggedResponse("APPEND saved-messages ",
                                       "$tag OK APPEND completed\r\n");

        var messageBody1 = @"Date: Mon, 7 Feb 1994 21:52:25 -0800 (PST)
From: Fred Foobar <foobar@Blurdybloop.example.COM>
Subject: afternoon meeting
To: mooch@owatagu.example.net
Message-Id: <B27397-0100000@Blurdybloop.example.COM>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

Hello Joe, do you think we can meet at 3:30 tomorrow?
".Replace("\r\n", "\n").Replace("\n", "\r\n");
        var messageBody2 = @"Date: Mon, 7 Feb 1994 22:43:04 -0800 (PST)
From: Joe Mooch <mooch@OWaTaGu.example.net>
Subject: Re: afternoon meeting
To: foobar@blurdybloop.example.com
Message-Id: <a0434793874930@OWaTaGu.example.net>
MIME-Version: 1.0
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

3:30 is fine with me.
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var message1 = new ImapAppendMessage(Encoding.ASCII.GetBytes(messageBody1),
                                             new ImapMessageFlagSet(ImapMessageFlag.Seen));
        var message2 = new ImapAppendMessage(Encoding.ASCII.GetBytes(messageBody2),
                                             new DateTimeOffset(new DateTime(1994, 2, 7, 22, 43, 04), new TimeSpan(-8, 0, 0)),
                                             new ImapMessageFlagSet(ImapMessageFlag.Seen));

        ImapAppendedUidSet appended;

        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages", out appended));

        var expected =
          string.Format("0004 APPEND saved-messages (\\Seen) {{{0}}}\r\n", Encoding.ASCII.GetByteCount(messageBody1)) +
          messageBody1 +
          string.Format(" (\\Seen) \"07-Feb-1994 22:43:04 -0800\" {{{0}}}\r\n", Encoding.ASCII.GetByteCount(messageBody2)) +
          messageBody2 +
          "\r\n";

        Assert.AreEqual(expected, server.DequeueRequest());

        if (respondAppendUid) {
          Assert.IsNotNull(appended);
          Assert.AreEqual(38505L, appended.UidValidity);

          var appendedUids = appended.ToArray();

          Assert.AreEqual(2, appendedUids.Length);
          Assert.AreEqual(3955L, appendedUids[0]);
          Assert.AreEqual(3956L, appendedUids[1]);
        }
      });
    }

    [Test]
    public void TestAppendMultiple_NonSynchronizedLiterals()
    {
      TestCases.WithMailboxSelectedSession(new[] {"MULTIAPPEND", "LITERAL+"}, (session, server) => {
        server.EnqueueTaggedResponse("$tag OK APPEND completed\r\n");

        var message1 = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message1"));
        var message2 = new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\ntest message2"));
        Assert.IsTrue((bool)session.AppendMultiple(new[] {message1, message2}, "saved-messages"));

        Assert.AreEqual("0004 APPEND saved-messages {34+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message1 {34+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "test message2\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppendMultiple_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(delegate {
          session.AppendMultiple(new[] {
            new ImapAppendMessage(Encoding.ASCII.GetBytes("message")),
            new ImapAppendMessage(Encoding.ASCII.GetBytes("message2")),
          }, "saved-messages");
        });
      });
    }

    [Test]
    public void TestAppendBinary()
    {
      TestCases.WithMailboxSelectedSession(new[] {"BINARY"}, (session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX",
                                     "$tag OK APPEND completed\r\n");

        Assert.IsTrue((bool)session.AppendBinary(new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\n\x00\x01\x02\x03\x04\x05\x06\x07")),
                                                 "INBOX"));

        Assert.AreEqual("0004 APPEND INBOX ~{29}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "\x00\x01\x02\x03\x04\x05\x06\x07\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestAppendBinary_NonSynchronizedLiterals()
    {
      TestCases.WithMailboxSelectedSession(new[] {"BINARY", "LITERAL+"}, (session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        Assert.IsTrue((bool)session.AppendBinary(new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\n\x00\x01\x02\x03\x04\x05\x06\x07")),
                                                 "INBOX"));

        Assert.AreEqual(("0004 APPEND INBOX ~{29+}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "\x00\x01\x02\x03\x04\x05\x06\x07\r\n").Replace("\x00", @"\0"),
                        server.DequeueRequest().Replace("\x00", @"\0"));
      });
    }

    [Test]
    public void TestAppendBinary_Multiple()
    {
      TestCases.WithMailboxSelectedSession(new[] {"BINARY", "MULTIAPPEND"}, (session, server) => {
        server.EnqueueTaggedResponse("APPEND INBOX ",
                                     "$tag OK APPEND completed\r\n");

        var messages = new[] {
          new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\nmessage1\x00\x01\x02\x03\x04\x05\x06\x07")),
          new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\nmessage2\x00\x01\x02\x03\x04\x05\x06\x07")),
        };

        Assert.IsTrue((bool)session.AppendBinaryMultiple(messages,
                                                         "INBOX"));

        Assert.AreEqual(("0004 APPEND INBOX ~{37}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "message1\x00\x01\x02\x03\x04\x05\x06\x07" +
                        " ~{37}\r\n" +
                        "MIME-Version: 1.0\r\n" +
                        "\r\n" +
                        "message2\x00\x01\x02\x03\x04\x05\x06\x07\r\n").Replace("\x00", @"\0"),
                        server.DequeueRequest().Replace("\x00", @"\0"));
      });
    }

    [Test]
    public void TestAppendBinary_Incapable()
    {
      TestCases.WithMailboxSelectedSession((session, server) => {
        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(delegate {
          session.AppendBinary(new ImapAppendMessage(Encoding.ASCII.GetBytes("MIME-Version: 1.0\r\n\r\n\x00\x01\x02\x03\x04\x05\x06\x07")),
                               "saved-messages");
        });
      });
    }
  }
}
