using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapClientTests {
    [Test]
    public void TestConstruct()
    {
      Assert.AreEqual("imap://localhost/",
                      DefaultPropertyAssertion(new ImapClient()).Profile.Authority.AbsoluteUri,
                      "#1 authority");

      Assert.AreEqual("imap://user;AUTH=LOGIN@imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient(new Uri("imap://user;AUTH=LOGIN@imap.example.net:10110/"))).Profile.Authority.AbsoluteUri,
                      "#2 authority");

      Assert.AreEqual("imap://imap.example.net/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net")).Profile.Authority.AbsoluteUri,
                      "#3 authority");

      Assert.AreEqual("imap://user@imap.example.net/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", "user")).Profile.Authority.AbsoluteUri,
                      "#4 authority");

      Assert.AreEqual("imaps://user@imap.example.net/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", true, "user")).Profile.Authority.AbsoluteUri,
                      "#5 authority");

      Assert.AreEqual("imap://imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110)).Profile.Authority.AbsoluteUri,
                      "#6 authority");

      Assert.AreEqual("imap://user@imap.example.net:10110/", 
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, "user")).Profile.Authority.AbsoluteUri,
                      "#7 authority");

      Assert.AreEqual("imap://user@imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, "user")).Profile.Authority.AbsoluteUri,
                      "#8 authority");

      Assert.AreEqual("imaps://user@imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, true, "user")).Profile.Authority.AbsoluteUri,
                      "#9 authority");

      Assert.AreEqual("imap://user;AUTH=PLAIN@imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, "user", "PLAIN")).Profile.Authority.AbsoluteUri,
                      "#10 authority");

      Assert.AreEqual("imaps://user;AUTH=DIGEST-MD5@imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, true, "user", "DIGEST-MD5")).Profile.Authority.AbsoluteUri,
                      "#11 authority");

      Assert.AreEqual("imaps://user@imap.example.net:10110/",
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, true, "user", null, 1000), 1000).Profile.Authority.AbsoluteUri,
                      "#12 authority");

      var profile = new ImapClientProfile(new Uri("imaps://user@imap.example.net/"));

      Assert.AreEqual("imaps://user@imap.example.net/",
                      DefaultPropertyAssertion(new ImapClient(profile)).Profile.Authority.AbsoluteUri,
                      "#13 authority");
    }

    private ImapClient DefaultPropertyAssertion(ImapClient client)
    {
      return DefaultPropertyAssertion(client, System.Threading.Timeout.Infinite);
    }

    private ImapClient DefaultPropertyAssertion(ImapClient client, int timeoutMilliseconds)
    {
      Assert.IsNotNull(client.Profile);
      Assert.IsTrue(client.Profile.UseTlsIfAvailable);
      Assert.IsFalse(client.Profile.AllowInsecureLogin);
      Assert.IsFalse(client.IsConnected);
      Assert.AreEqual(timeoutMilliseconds, client.Timeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.ReceiveTimeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.SendTimeout);
      Assert.IsFalse(client.IsBusy);

      Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.OpeningMailbox));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerID));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerNamespace));
      Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsIdling));

      Assert.DoesNotThrow(() => client.DisabledCapabilities.Clear());

      return client;
    }

    [Test]
    public void TestConnect_WithPassword()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnect_WithCredentials()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnect_WithSpecifiedSaslMechanism()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          client.Connect(authMechanism);

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    private class SaslXOAuth2 : XOAuth2Mechanism {
      public bool InitializeCalled = false;

      public SaslXOAuth2(string user, string accessToken)
        : base(user, accessToken)
      {
      }

      public override void Initialize()
      {
        base.Initialize();

        InitializeCalled = true;
      }
    }

    [Test]
    public void TestConnect_WithXOAuth2()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 SASL-IR] ImapPseudoServer ready\r\n"); // AUTH=XOAUTH2 not listed
        server.EnqueueTaggedResponse("AUTHENTICATE XOAUTH2 ",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        using (var oauth2 = new SaslXOAuth2("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
          client.Connect(oauth2);

          Assert.IsTrue(oauth2.InitializeCalled);
          Assert.AreEqual(SaslExchangeStatus.Continuing, oauth2.ExchangeStatus);
          Assert.IsNull(oauth2.LastErrorResults);
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
      });
    }

    [Test]
    public void TestConnect_WithXOAuth2_Error()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 SASL-IR] ImapPseudoServer ready\r\n"); // AUTH=XOAUTH2 not listed
        server.EnqueueResponse("AUTHENTICATE XOAUTH2 ",
                               "+ eyJzdGF0dXMiOiI0MDEiLCJzY2hlbWVzIjoiYmVhcmVyIG1hYyIsInNjb\n3BlIjoiaHR0cHM6Ly9tYWlsLmdvb2dsZS5jb20vIn0K\r\n");
        server.EnqueueTaggedResponse("$tag NO SASL authentication failed\r\n");

        using (var oauth2 = new SaslXOAuth2("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
          Assert.Throws<ImapAuthenticationException>(() => client.Connect(oauth2));

          Assert.IsTrue(oauth2.InitializeCalled);
          Assert.AreEqual(SaslExchangeStatus.Failed, oauth2.ExchangeStatus);
          Assert.AreEqual("{\"status\":\"401\",\"schemes\":\"bearer mac\",\"scope\":\"https://mail.google.com/\"}\n", 
                          oauth2.LastErrorResults);
        }

        Assert.IsFalse(client.IsConnected);
      });
    }

    [Test]
    public void TestConnect_AlreadyConnected()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

        Assert.Throws<InvalidOperationException>(() => client.Connect("pass"));

        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnect_IDCapable()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1 ID\r\n" + 
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("ID ",
                                     "* ID (\"name\" \"ImapPseudoServer\")\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.AreEqual("ImapPseudoServer", client.ServerID["name"]);
      });
    }

    [Test]
    public void TestConnect_NamespaceCapable()
    {
      Connect((server, client) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1 NAMESPACE\r\n" + 
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("NAMESPACE\r\n",
                                     "* NAMESPACE ((\"\" \"/\")) NIL NIL\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.AreEqual(1, client.ServerNamespace.PersonalNamespaces.Count);
        Assert.AreEqual("/", client.ServerNamespace.PersonalNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, client.ServerNamespace.SharedNamespaces.Count);
        Assert.AreEqual(0, client.ServerNamespace.OtherUsersNamespaces.Count);
      });
    }

    private void Connect(Action<ImapPseudoServer, ImapClient> action)
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          Assert.IsFalse(client.IsConnected);
          Assert.IsFalse(client.IsBusy);
          Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.OpeningMailbox));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerNamespace));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerID));

          action(server, client);

          Assert.IsFalse(client.IsBusy);

          if (client.IsConnected)
            Assert.IsFalse(client.IsIdling);
          else
            Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsIdling));
        }
      });
    }

#if DEBUG
    private const int testAsyncConnectRepetition = 3;
#else
    private const int testAsyncConnectRepetition = 20;
#endif

    [Test]
    public void TestBeginConnect_WithPassword()
    {
      TestUtils.Repeat(testAsyncConnectRepetition, () => BeginConnect(true));
    }

    [Test]
    public void TestBeginConnect_WithCredentials()
    {
      TestUtils.Repeat(testAsyncConnectRepetition, () => BeginConnect(false));
    }

    private void BeginConnect(bool usePassword)
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 5000;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          var callbacked = false;
          var asyncStateObject = new object();

          var asyncCallback = (AsyncCallback)delegate(IAsyncResult ar) {
            Assert.IsNotNull(ar);
            Assert.AreSame(ar.AsyncState, asyncStateObject);
            callbacked = true;
          };

          client.Profile.AllowInsecureLogin = true;

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          var asyncResult = usePassword
            ? client.BeginConnect("pass", asyncCallback, asyncStateObject)
            : client.BeginConnect(new NetworkCredential("user", "pass"), asyncCallback, asyncStateObject);

          Assert.IsFalse(client.IsConnected);
          Assert.IsNotNull(asyncResult);
          Assert.IsFalse(asyncResult.IsCompleted);
          Assert.IsFalse(callbacked);

          Assert.Throws<InvalidOperationException>(() => client.BeginConnect("pass"));

          Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

          var interval = TimeSpan.FromMilliseconds(100);
          Task responseTask = null;

          TestUtils.Assert.Elapses(interval, delegate {
            responseTask = TestCases.QueueResponseTask(server, client, delegate {
              Thread.Sleep(interval);

              server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                         "* CAPABILITY IMAP4rev1\r\n" +
                                         "$tag OK done\r\n");
            });

#if NETFRAMEWORK || NETSTANDARD2_0
            Assert.IsTrue(asyncResult.AsyncWaitHandle.WaitOne(3000, false));
#else
            Assert.IsTrue(asyncResult.AsyncWaitHandle.WaitOne(3000));
#endif
          });

          // Assert.IsFalse(client.IsConnected); // indefinite value

          Thread.Sleep(50); // wait for callback
          Assert.IsTrue(callbacked);
          //Assert.IsTrue(client.IsBusy, "IsBusy before EndConnect"); // indefinite value

          client.EndConnect(asyncResult);

          responseTask.Wait();

          Assert.IsFalse(client.IsBusy, "IsBusy after connected");

          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsSecureSession);

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          Assert.Throws<ArgumentException>(() => client.EndConnect(asyncResult));
        }
      });
    }

    [Test]
    public void TestBeginConnect_AlreadyConnected()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        Assert.Throws<InvalidOperationException>(() => client.BeginConnect("pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
      });
    }

    [Test]
    public void TestBeginConnect_Exception()
    {
      ImapPseudoServer.StartNotAccepting((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 500;

          Assert.IsFalse(client.IsConnected);

          var asyncResult = client.BeginConnect("pass");

          /*
           * no response
           */

          TestUtils.Assert.ThrowsOrAggregates<ImapConnectionException>(() => client.EndConnect(asyncResult));

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          Assert.IsFalse(client.IsConnected);
        }
      });
    }

    [Test]
    public void TestConnectAsync_WithPassword()
    {
      TestUtils.Repeat(testAsyncConnectRepetition, () => ConnectAsync(true));
    }
    
    [Test]
    public void TestConnectAsync_WithCredentials()
    {
      TestUtils.Repeat(testAsyncConnectRepetition, () => ConnectAsync(false));
    }
    
    private void ConnectAsync(bool usePassword)
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          var t = usePassword
            ? client.ConnectAsync("pass")
            : client.ConnectAsync(new NetworkCredential("user", "pass"));

          Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

          var interval = TimeSpan.FromMilliseconds(100);

          var responseTask = TestCases.QueueResponseTask(server, client, delegate {
            Thread.Sleep(interval);

            server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                         "* CAPABILITY IMAP4rev1\r\n" +
                                         "$tag OK done\r\n");
          });

          TestUtils.Assert.Elapses(interval, delegate {
            Assert.IsTrue(t.Wait(3000), "Task.Wait");
          });

          responseTask.Wait();

          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsBusy, "IsBusy after connected");
        }
      });
    }
    
    [Test]
    public void TestConnectAsync_AlreadyConnected()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        Assert.Throws<InvalidOperationException>(() => client.ConnectAsync("pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
      });
    }
    
    [Test]
    public void TestConnectAsync_Exception()
    {
      ImapPseudoServer.StartNotAccepting((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 250;

          Assert.IsFalse(client.IsConnected);

          var t = client.ConnectAsync("pass");

          var ex = Assert.Throws<AggregateException>(() => t.Wait(10000));

          Assert.IsInstanceOf<ImapConnectionException>(ex.InnerException);

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          Assert.IsFalse(client.IsConnected);
        }
      });
    }

    [Test]
    public void TestConnect_CancellationRequested()
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 1000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          var cts = new CancellationTokenSource();

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n",
                                     () => cts.Cancel(),
                                     null);

          Assert.Throws<OperationCanceledException>(() => client.Connect("pass", null, cts.Token));

          Assert.IsFalse(client.IsBusy, "IsBusy after cancel");
          Assert.IsFalse(client.IsConnected);
          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
        }
      });
    }

    [Test]
    public void TestConnectAsync_CancellationRequested()
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 1000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          using (var e = new ManualResetEventSlim(false)) {
            var cts = new CancellationTokenSource();

            server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n",
                                       () => { e.Wait(); cts.Cancel(); },
                                       null);

            var t = client.ConnectAsync("pass", null, cts.Token);

            Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

            e.Set();

            Assert.Throws<AggregateException>(() => t.Wait());

            Assert.IsTrue(t.IsCanceled, "IsCanceled");
          }

          Assert.IsFalse(client.IsBusy, "IsBusy after cancel");
          Assert.IsFalse(client.IsConnected);
          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
        }
      });
    }

    [Test]
    public void TestConnect_SecurePortWithDefaultUpgradeStreamCallback()
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, true, "user")) {
          Assert.IsFalse(client.IsConnected);

          Assert.Throws<ImapSecureConnectionException>(() => client.Connect("pass"));

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          Assert.IsFalse(client.IsConnected);
        }
      });
    }

    [Test]
    public void TestConnectAsync_SecurePortWithDefaultUpgradeStreamCallback()
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, true, "user")) {
          Assert.IsFalse(client.IsConnected);

          var t = client.ConnectAsync("pass");

          Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

          TestUtils.Assert.ThrowsOrAggregates<ImapSecureConnectionException>(() => t.Wait());

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          Assert.IsFalse(client.IsBusy, "IsBusy after exception");
          Assert.IsFalse(client.IsConnected);
        }
      });
    }

    [Test]
    public void TestLogout_BeginConnectRunning()
    {
      Logout_AsyncConnectRunning(client => client.BeginConnect("pass"),
                                 (client, ar) => client.EndConnect(ar));
    }

    [Test]
    public void TestLogout_ConnectAsyncRunning()
    {
      Logout_AsyncConnectRunning(client => client.ConnectAsync("pass"),
                                 (client, task) => task.Wait());
    }

    [Test]
    public void TestDisconnect_BeginConnectRunning()
    {
      Logout_AsyncConnectRunning(client => client.BeginConnect("pass"),
                                 (client, ar) => client.EndConnect(ar));
    }

    [Test]
    public void TestDisconnect_ConnectAsyncRunning()
    {
      Logout_AsyncConnectRunning(client => client.ConnectAsync("pass"),
                                 (client, task) => task.Wait());
    }

    private void Logout_AsyncConnectRunning<TAsyncResult>(Func<ImapClient, TAsyncResult> startConnect,
                                                          Action<ImapClient, TAsyncResult> endConnect)
    {
      LogoutDisconnect_AsyncConnectRunning(startConnect, endConnect, client => client.Logout());
    }

    private void Disconnect_AsyncConnectRunning<TAsyncResult>(Func<ImapClient, TAsyncResult> startConnect,
                                                              Action<ImapClient, TAsyncResult> endConnect)
    {
      LogoutDisconnect_AsyncConnectRunning(startConnect, endConnect, client => client.Disconnect());
    }

    private void LogoutDisconnect_AsyncConnectRunning<TAsyncResult>(Func<ImapClient, TAsyncResult> startConnect,
                                                                    Action<ImapClient, TAsyncResult> endConnect,
                                                                    Action<ImapClient> testAction)
    {
      ImapPseudoServer.Start(false, (server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.AllowInsecureLogin = true;
          client.Profile.Timeout = 5000;

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          var asyncResult = startConnect(client);

          Assert.IsTrue(client.IsBusy);

          Assert.Throws<InvalidOperationException>(() => testAction(client));

          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK done\r\n");

          endConnect(client, asyncResult);

          Assert.IsFalse(client.IsBusy);
          Assert.IsTrue(client.IsConnected);
        }
      });
    }

    [Test]
    public void TestDispose_BeginConnectRunning()
    {
      Dispose_AsyncConnectRunning(client => client.BeginConnect("pass"),
                                  (client, ar) => client.EndConnect(ar));
    }

    [Test]
    public void TestDispose_ConnectAsyncRunning()
    {
      Dispose_AsyncConnectRunning(client => client.ConnectAsync("pass"),
                                  (client, task) => task.Wait());
    }

    private void Dispose_AsyncConnectRunning<TAsyncResult>(Func<ImapClient, TAsyncResult> startConnect,
                                                           Action<ImapClient, TAsyncResult> endConnect)
    {
      ImapPseudoServer.Start((server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.AllowInsecureLogin = true;
          client.Profile.Timeout = 5000;

          Assert.IsFalse(client.IsConnected, "IsConnected before BeginConnect");

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          var asyncResult = startConnect(client);

          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" +
                                       "$tag OK done\r\n");

          ((IDisposable)client).Dispose();

          Assert.IsFalse(client.IsConnected, "IsConnected after Dispose");

          endConnect(client, asyncResult);

          Assert.IsFalse(client.IsConnected, "IsConnected after EndConnect");
        }
      });
    }

    [Test]
    public void TestEndConnect_InvalidAsyncResult()
    {
      ImapPseudoServer.Start(false, (server1) => {
        ImapPseudoServer.Start(false, (server2) => {
          using (ImapClient
                 client1 = new ImapClient(server1.Host, server1.Port, "user1"),
                 client2 = new ImapClient(server2.Host, server2.Port, "user2")) {
            client1.Profile.ReceiveTimeout = 100;
            client2.Profile.ReceiveTimeout = 100;

            var asyncResult1 = client1.BeginConnect("pass1");
            var asyncResult2 = client2.BeginConnect(new NetworkCredential("user2", "pass2"));

            Assert.Throws<ArgumentException>(() => client1.EndConnect(asyncResult2));
            Assert.Throws<ArgumentException>(() => client2.EndConnect(asyncResult1));

            try {
              client1.EndConnect(asyncResult1);
            }
            catch (AggregateException ex) {
              if (!(ex.GetBaseException() is TimeoutException))
                throw;

              // possible exception
            }
            catch (TimeoutException) {
              // possible exception
            }

            try {
              client2.EndConnect(asyncResult2);
            }
            catch (AggregateException ex) {
              if (!(ex.GetBaseException() is TimeoutException))
                throw;

              // possible exception
            }
            catch (TimeoutException) {
              // possible exception
            }
          }
        });
      });
    }

    [Test]
    public void TestConnect_BeginConnectRunning()
    {
      ImapPseudoServer.Start(false, (server) => {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.BeginConnect("pass");

          Assert.Throws<InvalidOperationException>(() => client.Connect("pass"));

          Assert.Throws<InvalidOperationException>(() => client.Connect(new NetworkCredential("user", "pass")));
        }
      });
    }

    [Test]
    public void TestDisconnect()
    {
      DisposeOrDisconnect(true);
    }

    [Test]
    public void TestDispose()
    {
      DisposeOrDisconnect(false);
    }

    private void DisposeOrDisconnect(bool disconnect)
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        if (disconnect)
          client.Disconnect();
        else
          ((IDisposable)client).Dispose();

        DisconnectedPropertyAssertion(client);

        if (disconnect)
          Assert.DoesNotThrow(() => client.Disconnect(), "Disconnect again");
        else
          Assert.DoesNotThrow(() => ((IDisposable)client).Dispose(), "Dispose again");
      });
    }

    private void DisconnectedPropertyAssertion(ImapClient client)
    {
      Assert.IsFalse(client.IsConnected);
      Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.OpeningMailbox));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerNamespace));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerID));
    }

    [Test]
    public void TestCheckConnected()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        client.Disconnect();

        Assert.IsFalse(client.IsConnected);

        Assert.Throws<InvalidOperationException>(() => client.Refresh());
      });
    }

    [Test]
    public void TestOperationAfterDisconnected()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        client.Disconnect();

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestOperationAfterTimeout()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        client.Timeout = 100;

        try {
          client.Refresh();
        }
        catch (TimeoutException) {
        }

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestOperationAfterInternalError()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueResponse("-ERR\r\n");

        try {
          client.Refresh();
        }
        catch (ImapException) {
        }

        OperationAfterException(client);
      });
    }

    private void OperationAfterException(ImapClient client)
    {
      Assert.Throws<InvalidOperationException>(() => client.Refresh());

      DisconnectedPropertyAssertion(client);
    }

    [Test]
    public void TestDisconnectedFromServerWithByeResponse()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueUntaggedByeResponse("* BYE server shutdown\r\n");

        var ex = Assert.Throws<ImapConnectionException>(() => client.Refresh());

        StringAssert.Contains("server shutdown", ex.Message);

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestDisconnectedFromServerWithoutByeResponse()
    {
      ImapPseudoServer.Start((server) => {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n",
                                     true);

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;
          client.Connect("pass");

          Assert.Throws<ImapConnectionException>(() => client.Refresh());

          OperationAfterException(client);
        }
      });
    }

    [Test]
    public void TestLogout()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK done\r\n",
                                     true);

        client.Logout();

        DisconnectedPropertyAssertion(client);

        Assert.DoesNotThrow(() => client.Logout(), "Logout again");
      });
    }

    [Test]
    public void TestLogout_InSelectedState()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK done\r\n");

        Assert.IsNotNull(client.OpenInbox());
        Assert.IsNotNull(client.OpeningMailbox);
        Assert.IsTrue(client.OpeningMailbox.IsInbox);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK done\r\n",
                                     true);

        client.Logout();

        DisconnectedPropertyAssertion(client);
      });
    }

    [Test]
    public void TestConnect_Reconnect()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK done\r\n",
                                     true);

        client.Logout();

        Assert.IsFalse(client.IsConnected);

        // reconnect
        server.Restart();

        client.Profile.Host = server.Host;
        client.Profile.Port = server.Port;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
      });
    }

    [Test]
    public void TestSerializeBinary_ErrorResponseException()
    {
      TestCases.WithAuthenticatedClient(new[] {ImapCapability.Quota}, (server, client) => {
        server.EnqueueTaggedResponse("GETQUOTA ",
                                     "$tag NO no such quota root\r\n");

        var ex = Assert.Throws<ImapErrorResponseException>(() => client.GetQuotaUsage("NON-EXISTENT", "STORAGE"));

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapErrorResponseException deserialized) {
          Assert.AreEqual(ex.Message, deserialized.Message);
          Assert.IsNotNull(deserialized.Result);
          Assert.IsNotNull(deserialized.Result.TaggedStatusResponse);
          Assert.AreEqual(ImapResponseCondition.No, deserialized.Result.TaggedStatusResponse.Condition);
          Assert.AreEqual("no such quota root", deserialized.Result.ResponseText);

          var responses = deserialized.Result.ReceivedResponses.ToArray();

          Assert.AreEqual(1, responses.Length);
          Assert.IsInstanceOf<ImapTaggedStatusResponse>(responses[0]);
        });
      });
    }
      
    [Test]
    public void TestInternalImpl_RunExclusively()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        TestUtils.Repeat(10, () => {
          var responded = 0;
          var responseTask = TestCases.QueueResponseTask(server, client, delegate {
            Thread.Sleep(100);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n",
                                         () => Interlocked.Exchange(ref responded, 1),
                                         null);
          });

          var t = Task.Run(() => client.Refresh()); // call RunExclusively()

          server.WaitForClientRequest(); // wait for Refresh() started

          var count = 0;

          for (;;) {
            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.Refresh());

            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.RefreshAsync());

            count++;

            Thread.Sleep(1);
          }

          Assert.DoesNotThrow(() => t.Wait());

          Assert.Greater(count, 0, "count of method call");

          responseTask.Wait();
        });
      });
    }

    [Test]
    public void TestInternalImpl_StartTaskExclusively()
    {
      TestCases.WithAuthenticatedClient((server, client) => {
        TestUtils.Repeat(10, () => {
          var responded = 0;
          var responseTask = TestCases.QueueResponseTask(server, client, delegate {
            Thread.Sleep(100);

            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n",
                                         () => Interlocked.Exchange(ref responded, 1),
                                         null);
          });

          var t = client.RefreshAsync(); // call StartTaskExclusively()

          var count = 0;

          for (;;) {
            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.Refresh());

            if (1 == Interlocked.CompareExchange(ref responded, 1, 1))
              break;

            Assert.Throws<InvalidOperationException>(() => client.RefreshAsync());

            count++;

            Thread.Sleep(1);
          }

          Assert.DoesNotThrow(() => t.Wait());

          Assert.Greater(count, 0, "count of method call");

          responseTask.Wait();
        });
      });
    }
  }
}
