using System;
using System.Linq;
using System.Text;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessageInfoBaseTests {
    private class MessageSet : ImapMessageInfoBase {
      public MessageSet(ImapOpenedMailboxInfo mailbox, ImapSequenceSet sequenceOrUidSet)
        : base(mailbox)
      {
        this.sequenceOrUidSet = sequenceOrUidSet;
      }

      protected override ImapSequenceSet GetSequenceOrUidSet()
      {
        return sequenceOrUidSet;
      }

      private ImapSequenceSet sequenceOrUidSet;
    }

    private void TestMessageSet(int existMesasgeCount,
                                ImapSequenceSet sequenceOrUidSet,
                                Action<ImapPseudoServer, ImapMessageInfoBase, ImapMessageInfo[]> action)
    {
      TestMessageSet(existMesasgeCount, null, sequenceOrUidSet, action);
    }

    private void TestMessageSet(int existMessageCount,
                                ImapCapability[] capabilities,
                                ImapSequenceSet sequenceOrUidSet,
                                Action<ImapPseudoServer, ImapMessageInfoBase, ImapMessageInfo[]> action)
    {
      TestCases.WithAuthenticatedClient(capabilities, (server, client) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \".\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     string.Format("* EXISTS {0}\r\n", existMessageCount) +
                                     "* OK [UIDVALIDITY 1]\r\n" +
                                     "$tag OK done\r\n");

        using (var inbox = client.OpenInbox()) {
          server.DequeueRequest(); // LIST
          server.DequeueRequest(); // SELECT

          var messages = new ImapMessageInfo[0];

          if (0 < existMessageCount) {
            server.EnqueueTaggedResponse("NOOP\r\n",
                                         "$tag OK done\r\n");

            var fetchResp = new StringBuilder();

            for (var seq = 1; seq <= existMessageCount; seq++) {
              fetchResp.AppendFormat("* FETCH {0} (UID {0})\r\n", seq);
            }

            fetchResp.Append("$tag OK done\r\n");

            server.EnqueueTaggedResponse("FETCH ",
                                         fetchResp.ToString());

            var msgs = inbox.GetMessages();

            messages = msgs.ToArray();

            server.DequeueRequest(); // NOOP
            server.DequeueRequest(); // FETCH

            Assert.AreSame(client, msgs.Client);
            Assert.AreSame(inbox, msgs.Mailbox);
            Assert.IsTrue(msgs.IsAvailable);
          }

          try {
            var messageSet = new MessageSet(inbox, sequenceOrUidSet);

            Assert.AreSame(client, messageSet.Client);
            Assert.AreSame(inbox, messageSet.Mailbox);
            Assert.IsTrue(messageSet.IsAvailable);
            Assert.AreEqual(1L, messageSet.UidValidity);

            action(server, messageSet, messages);
          }
          finally {
            if (inbox.IsOpen)
              server.EnqueueTaggedResponse("CLOSE\r\n",
                                           "$tag OK done\r\n");
          }
        }
      });
    }

    private ImapSequenceSet CreateEmptySet()
    {
      return ImapSequenceSet.CreateSet(new long[0]);
    }

    [Test]
    public void TestAddFlags()
    {
      AddFlags((messages, f) => messages.AddFlags(f));
    }

    [Test]
    public void TestAddFlagsAsync()
    {
      AddFlags((messages, f) => messages.AddFlagsAsync(f).Wait());
    }

    private void AddFlags(Action<ImapMessageInfoBase, ImapMessageFlag> testAction)
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Draft)\r\n",
                                     "* FETCH 1 (FLAGS (\\Seen \\Draft))\r\n" +
                                     "$tag OK done\r\n");
        });

        testAction(messageSet, ImapMessageFlag.Draft);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].IsSeen);
        Assert.IsTrue(messages[0].IsDraft);
      });
    }

    [Test]
    public void TestRemoveFlags()
    {
      RemoveFlags((messages, f) => messages.RemoveFlags(f));
    }

    [Test]
    public void TestRemoveFlagsAsync()
    {
      RemoveFlags((messages, f) => messages.RemoveFlagsAsync(f).Wait());
    }

    private void RemoveFlags(Action<ImapMessageInfoBase, ImapMessageFlag> testAction)
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(messageSet, ImapMessageFlag.Draft);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);
      });
    }

    [Test]
    public void TestSetFlags()
    {
      SetFlags((messages, f1, f2) => messages.SetFlags(f1, f2));
    }

    [Test]
    public void TestSetFlagsAsync()
    {
      SetFlags((messages, f1, f2) => messages.SetFlagsAsync(f1, f2).Wait());
    }

    private void SetFlags(Action<ImapMessageInfoBase, ImapMessageFlag, ImapMessageFlag> testAction)
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 FLAGS (\\Draft \\Flagged)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft \\Flagged))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(messageSet, ImapMessageFlag.Draft, ImapMessageFlag.Flagged);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(messages[0].IsDraft);
        Assert.IsTrue(messages[0].IsFlagged);
      });
    }

    [Test]
    public void TestAddFlags_NonApplicableFlagsRemoved()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.AddFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestRemoveFlags_NonApplicableFlagsRemoved()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 -FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft \\Recent))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.RemoveFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(2, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestSetFlags_NonApplicableFlagsRemoved()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 FLAGS (\\Draft)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.SetFlags(ImapMessageFlag.Draft, ImapMessageFlag.Recent);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsFalse(messages[0].Flags.Contains(ImapMessageFlag.Recent));
      });
    }

    [Test]
    public void TestAddFlags_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => messageSet.AddFlags(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestAddFlagsAsync_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => messageSet.AddFlagsAsync(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestRemoveFlags_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => messageSet.RemoveFlags(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestRemoveFlagsAsync_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => messageSet.RemoveFlagsAsync(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestSetFlags_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => messageSet.SetFlags(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestSetFlagsAsync_ContainsNoApplicableFlags()
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => messageSet.SetFlagsAsync(ImapMessageFlag.Recent));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestStore_EmptySet()
    {
      Store_EmptySet(messages => messages.Store(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen)));
    }

    [Test]
    public void TestStoreAsync_EmptySet()
    {
      Store_EmptySet(messages => messages.StoreAsync(ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen)).Wait());
    }

    private void Store_EmptySet(Action<ImapMessageInfoBase> testAction)
    {
      TestMessageSet(0, CreateEmptySet(),
                     (server, messageSet, messages) => {
        // do nothing
        testAction(messageSet);

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestStore_SingleUidSet()
    {
      Store_SingleUidSet((messages, item) => messages.Store(item));
    }

    [Test]
    public void TestStoreAsync_SingleUidSet()
    {
      Store_SingleUidSet((messages, item) => messages.StoreAsync(item).Wait());
    }

    private void Store_SingleUidSet(Action<ImapMessageInfoBase, ImapStoreDataItem> testAction)
    {
      TestMessageSet(1, ImapSequenceSet.CreateUidSet(1L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID STORE 1 +FLAGS (\\Seen)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(messageSet, ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);
      });
    }

    [Test]
    public void TestStore_SequenceRangeSet()
    {
      Store_SingleUidSet((messages, item) => messages.Store(item));
    }

    [Test]
    public void TestStoreAsync_SequenceRangeSet()
    {
      Store_SingleUidSet((messages, item) => messages.StoreAsync(item).Wait());
    }

    private void Store_SequenceRangeSet(Action<ImapMessageInfoBase, ImapStoreDataItem> testAction)
    {
      TestMessageSet(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("STORE 1:3 -FLAGS (\\Seen)\r\n",
                                       "* FETCH 1 (FLAGS (\\Draft))\r\n" +
                                       "* FETCH 2 (FLAGS ($label1))\r\n" +
                                       "* FETCH 3 (FLAGS ())\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(messageSet, ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[0].IsDraft);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(1, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains("$label1"));

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(0, messages[2].Flags.Count);
      });
    }

    [Test]
    public void TestStore_MailboxClosed()
    {
      Store_MailboxClosed((messages, item) => messages.Store(item));
    }

    [Test]
    public void TestStoreAsync_MailboxClosed()
    {
      Store_MailboxClosed((messages, item) => messages.StoreAsync(item).Wait());
    }

    private void Store_MailboxClosed(Action<ImapMessageInfoBase, ImapStoreDataItem> testAction)
    {
      TestMessageSet(10, ImapSequenceSet.CreateRangeSet(1L, 10L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        messageSet.Mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);

        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapUnavailableException>(() => testAction(messageSet, ImapStoreDataItem.RemoveFlags(ImapMessageFlag.Seen)));

        Assert.IsFalse(messageSet.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestStore_UidValidityChanged()
    {
    }

    [Test]
    public void TestMarkAsDeleted()
    {
      MarkAsDeleted(messages => messages.MarkAsDeleted());
    }

    [Test]
    public void TestMarkAsDeletedAsync()
    {
      MarkAsDeleted(messages => messages.MarkAsDeletedAsync().Wait());
    }

    private void MarkAsDeleted(Action<ImapMessageInfoBase> testAction)
    {
      TestMessageSet(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(messageSet);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messages[0].IsAvailable);
        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsTrue(messages[1].IsAvailable);
        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);

        Assert.IsTrue(messages[2].IsAvailable);
        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMarkAsSeen()
    {
      MarkAsSeen(messages => messages.MarkAsSeen());
    }

    [Test]
    public void TestMarkAsSeenAsync()
    {
      MarkAsSeen(messages => messages.MarkAsSeenAsync().Wait());
    }

    private void MarkAsSeen(Action<ImapMessageInfoBase> testAction)
    {
      TestMessageSet(3, ImapSequenceSet.CreateRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Seen)\r\n",
                                       "* FETCH 1 (FLAGS (\\Seen))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Seen \\Draft))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Seen $label1))\r\n" +
                                       "$tag OK done\r\n");
        });

        testAction(messageSet);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[0].IsSeen);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(messages[1].IsSeen);
        Assert.IsTrue(messages[1].IsDraft);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsSeen);
      });
    }

    [Test]
    public void TestDelete_UidPlusCapable()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.UidPlus},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        Delete(server, messageSet, messages, true, m => m.Delete());
      });
    }

    [Test]
    public void TestDelete_UidPlusIncapable()
    {
      TestMessageSet(3,
                     null,
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        Delete(server, messageSet, messages, false, m => m.Delete());
      });
    }

    [Test]
    public void TestDeleteAsync_UidPlusCapable()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.UidPlus},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        Delete(server, messageSet, messages, true, m => m.DeleteAsync().Wait());
      });
    }

    [Test]
    public void TestDeleteAsync_UidPlusIncapable()
    {
      TestMessageSet(3,
                     null,
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        Delete(server, messageSet, messages, false, m => m.DeleteAsync().Wait());
      });
    }

    [Test]
    public void TestDelete_UidPlusDisabled()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.UidPlus},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.UidPlus);

        Delete(server, messageSet, messages, false, m => m.Delete());
      });
    }

    private void Delete(ImapPseudoServer server,
                        ImapMessageInfoBase messageSet,
                        ImapMessageInfo[] messages,
                        bool uidPlusAvailable,
                        Action<ImapMessageInfoBase> testAction)
    {
      var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                     "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                     "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                     "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse(uidPlusAvailable ? "UID EXPUNGE 1:3\r\n" : "EXPUNGE\r\n",
                                     "* EXPUNGE 3\r\n" +
                                     "* EXPUNGE 2\r\n" +
                                     "* EXPUNGE 1\r\n" +
                                     "$tag OK done\r\n");
      });

      var mailbox = messageSet.Mailbox;

      testAction(messageSet);

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      Assert.IsFalse(messageSet.IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
      Assert.AreEqual(0L, mailbox.ExistMessageCount);

      Assert.IsFalse(messages[0].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[0].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
      Assert.IsTrue(messages[0].IsDeleted);

      Assert.IsNotNull(messages[0].Flags);
      Assert.AreEqual(1, messages[0].Flags.Count);
      Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[0].IsMarkedAsDeleted);

      Assert.IsFalse(messages[1].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[1].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
      Assert.IsTrue(messages[1].IsDeleted);

      Assert.IsNotNull(messages[1].Flags);
      Assert.AreEqual(2, messages[1].Flags.Count);
      Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
      Assert.IsTrue(messages[1].IsMarkedAsDeleted);
      Assert.IsTrue(messages[1].IsSeen);

      Assert.IsFalse(messages[0].IsAvailable);
      Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messages[2].Mailbox));
      Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
      Assert.IsTrue(messages[2].IsDeleted);

      Assert.IsNotNull(messages[2].Flags);
      Assert.AreEqual(2, messages[2].Flags.Count);
      Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
      Assert.IsTrue(messages[2].Flags.Contains("$label1"));
      Assert.IsTrue(messages[2].IsMarkedAsDeleted);
    }

    [Test]
    public void TestDelete_EmptySet()
    {
      Delete_EmptySet(messages => messages.Delete());
    }

    [Test]
    public void TestDeleteAsync_EmptySet()
    {
      Delete_EmptySet(messages => messages.DeleteAsync().Wait());
    }

    private void Delete_EmptySet(Action<ImapMessageInfoBase> testAction)
    {
      TestMessageSet(0, CreateEmptySet(),
                     (server, messageSet, messages) => {
        // do nothing
        testAction(messageSet);

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
        Assert.DoesNotThrow(() => Assert.IsNotNull(messageSet.Mailbox));
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, false, (m, dest) => m.CopyTo(dest));
      });
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, false, (m, dest) => m.CopyToAsync(dest).Wait());
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true, (m, dest) => m.MoveTo(dest));
      });
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true, (m, dest) => m.MoveToAsync(dest).Wait());
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MoveDisabled()
    {
      TestMessageSet(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.Move);

        CopyOrMoveToDestinationMailbox(server, messageSet, messages, true, (m, dest) => m.MoveTo(dest));
      });
    }

    private void CopyOrMoveToDestinationMailbox(ImapPseudoServer server,
                                                ImapMessageInfoBase messageSet,
                                                ImapMessageInfo[] messages,
                                                bool move,
                                                Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      server.EnqueueTaggedResponse("LIST ",
                                   "* LIST () \"\" dest\r\n" +
                                   "$tag OK done\r\n");

      var destMailbox = messageSet.Client.GetMailbox("dest");

      var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("UID COPY 1,3,5 dest\r\n",
                                     "$tag OK done\r\n");
        if (move)
          server.EnqueueTaggedResponse("UID STORE 1,3,5 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
      });

      testAction(messageSet, destMailbox);

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      if (move) {
        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }
    
    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox()
    {
      TestMessageSet(3,
                     new[] {ImapCapability.Move},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var sourceMailbox = messageSet.Mailbox;
        var destMailbox = messageSet.Client.GetMailbox("dest");

        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                       "* EXPUNGE 3\r\n" +
                                       "* EXPUNGE 2\r\n" +
                                       "* EXPUNGE 1\r\n" +
                                       "$tag OK done\r\n");
        });

        messageSet.MoveTo(destMailbox);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
        Assert.AreEqual(0L, sourceMailbox.ExistMessageCount);

        Assert.IsFalse(messages[0].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.IsFalse(messages[1].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.IsFalse(messages[2].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_CopyFailed()
    {
      MoveToDestinationMailboxCopyFailed(false);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_CopyFailed()
    {
      MoveToDestinationMailboxCopyFailed(true);
    }

    private void MoveToDestinationMailboxCopyFailed(bool moveCapable)
    {
      var capabilities = moveCapable ? new[] { ImapCapability.Move } : Array.Empty<ImapCapability>();

      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse(moveCapable ? "UID MOVE 1,3,5 dest\r\n" : "UID COPY 1,3,5 dest\r\n",
                                       "$tag NO failed\r\n");
        });

        // STORE won't be issued
        Assert.Throws<ImapErrorResponseException>(() => messageSet.MoveTo(destMailbox));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(Array.Empty<ImapCapability>(),
                                           false,
                                           (messages, dest) => messages.CopyTo(dest));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(Array.Empty<ImapCapability>(),
                                           false,
                                           (messages, dest) => messages.CopyToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(Array.Empty<ImapCapability>(),
                                           true,
                                           (messages, dest) => messages.MoveTo(dest));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(Array.Empty<ImapCapability>(),
                                           true,
                                           (messages, dest) => messages.MoveToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_MailboxItself()
    {
      CopyOrMoveToDestinationMailboxItself(new[] {ImapCapability.Move},
                                           true,
                                           (messages, dest) => messages.MoveTo(dest));
    }

    private void CopyOrMoveToDestinationMailboxItself(ImapCapability[] capabilities,
                                                      bool move,
                                                      Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
        Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);

        if (move) {
          // do nothing
          testAction(messageSet, messageSet.Mailbox);
        }
        else {
          var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
            server.EnqueueTaggedResponse("UID COPY 1,3,5 " + messageSet.Mailbox.FullName + "\r\n",
                                         "* 3 RECENT\r\n" +
                                         "* 6 EXISTS\r\n" +
                                         "$tag OK done\r\n");
          });

          testAction(messageSet, messageSet.Mailbox);

          responseTask.Wait();
        }

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);

        if (move) {
          Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);
        }
        else {
          Assert.AreEqual(6, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(3, messageSet.Mailbox.RecentMessageCount);
        }
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(Array.Empty<ImapCapability>(),
                                                       (messages, dest) => messages.CopyTo(dest));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(Array.Empty<ImapCapability>(),
                                                       (messages, dest) => messages.CopyToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(Array.Empty<ImapCapability>(),
                                                       (messages, dest) => messages.MoveTo(dest));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(Array.Empty<ImapCapability>(),
                                                       (messages, dest) => messages.MoveToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      CopyOrMoveToDestinationMailboxOfDifferentSession(new[] {ImapCapability.Move},
                                                       (messages, dest) => messages.MoveTo(dest));
    }

    private void CopyOrMoveToDestinationMailboxOfDifferentSession(ImapCapability[] capabilities,
                                                                  Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (serverSrc, srcMessageSet, srcMessages) => {
        TestCases.WithAuthenticatedClient((serverDest, clientDest) => {
          serverDest.EnqueueTaggedResponse("LIST ",
                                           "* LIST () \"\" dest\r\n" +
                                           "$tag OK done\r\n");

          var destMailbox = clientDest.GetMailbox("dest");

          Assert.AreNotSame(srcMessageSet.Client, destMailbox.Client);

          var responseTask = TestCases.QueueResponseTask(serverSrc, srcMessageSet.Client, delegate {
            serverSrc.EnqueueTaggedResponse("UID FETCH 1,3,5 (UID)\r\n",
                                            "$tag NO FETCH failed\r\n");
          });

          TestUtils.Assert.ThrowsOrAggregates<ImapErrorResponseException>(() => testAction(srcMessageSet, destMailbox));

          responseTask.Wait();

          Assert.IsFalse(srcMessageSet.Client.IsBusy);
        });
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestCopyTo_CopyToDestinationMailbox_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveToDestinationMailbox_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_UidValidityChanged()
    {
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_ArgumentNull()
    {
      TestMessageSet(3, Array.Empty<ImapCapability>(), ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentNullException>(() => messageSet.CopyTo((ImapMailboxInfo)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_ArgumentNull()
    {
      TestMessageSet(3, Array.Empty<ImapCapability>(), ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentNullException>(() => messageSet.CopyToAsync((ImapMailboxInfo)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_ArgumentNull()
    {
      TestMessageSet(3, Array.Empty<ImapCapability>(), ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentNullException>(() => messageSet.MoveTo((ImapMailboxInfo)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_ArgumentNull()
    {
      TestMessageSet(3, Array.Empty<ImapCapability>(), ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentNullException>(() => messageSet.MoveToAsync((ImapMailboxInfo)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_ArgumentNull()
    {
      TestMessageSet(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentNullException>(() => messageSet.MoveTo((ImapMailboxInfo)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(Array.Empty<ImapCapability>(),
                                                (messages, dest) => messages.CopyTo(dest));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(Array.Empty<ImapCapability>(),
                                                (messages, dest) => messages.CopyToAsync(dest));
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(Array.Empty<ImapCapability>(),
                                                (messages, dest) => messages.MoveTo(dest));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(Array.Empty<ImapCapability>(),
                                                (messages, dest) => messages.MoveToAsync(dest));
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_NonExistent()
    {
      CopyOrMoveToDestinationMailboxNonExistent(new[] {ImapCapability.Move},
                                                (messages, dest) => messages.MoveTo(dest));
    }

    private void CopyOrMoveToDestinationMailboxNonExistent(ImapCapability[] capabilities,
                                                           Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      TestMessageSet(10,
                     capabilities.Prepend(ImapCapability.ListExtended),
                     ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST (\\Subscribed \\NonExistent) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        Assert.IsFalse(destMailbox.Exists);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(messageSet, destMailbox));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(Array.Empty<ImapCapability>(),
                                                 (messages, dest) => messages.CopyTo(dest));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(Array.Empty<ImapCapability>(),
                                                 (messages, dest) => messages.CopyToAsync(dest));
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(Array.Empty<ImapCapability>(),
                                                 (messages, dest) => messages.MoveTo(dest));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(Array.Empty<ImapCapability>(),
                                                 (messages, dest) => messages.MoveToAsync(dest));
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_Unselectable()
    {
      CopyOrMoveToDestinationMailboxUnselectable(new[] {ImapCapability.Move},
                                                 (messages, dest) => messages.MoveTo(dest));
    }

    private void CopyOrMoveToDestinationMailboxUnselectable(ImapCapability[] capabilities,
                                                            Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      TestMessageSet(10,
                     capabilities.Prepend(ImapCapability.ListExtended),
                     ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST (\\Noselect) \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest", ImapMailboxListOptions.SubscribedOnly);

        server.DequeueRequest();

        Assert.IsTrue(destMailbox.IsUnselectable);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(messageSet, destMailbox));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(Array.Empty<ImapCapability>(),
                                             (messages, dest) => messages.CopyTo(dest));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(Array.Empty<ImapCapability>(),
                                             (messages, dest) => messages.CopyToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(Array.Empty<ImapCapability>(),
                                             (messages, dest) => messages.MoveTo(dest));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(Array.Empty<ImapCapability>(),
                                             (messages, dest) => messages.MoveToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_EmptySet()
    {
      CopyOrMoveToDestinationMailboxEmptySet(new[] {ImapCapability.Move},
                                             (messages, dest) => messages.MoveTo(dest));
    }

    private void CopyOrMoveToDestinationMailboxEmptySet(ImapCapability[] capabilities,
                                                        Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      TestMessageSet(0, capabilities, CreateEmptySet(),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        // do nothing
        testAction(messageSet, destMailbox);

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(Array.Empty<ImapCapability>(),
                                                  (messages, dest) => messages.CopyTo(dest));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(Array.Empty<ImapCapability>(),
                                                  (messages, dest) => messages.CopyToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(Array.Empty<ImapCapability>(),
                                                  (messages, dest) => messages.MoveTo(dest));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(Array.Empty<ImapCapability>(),
                                                  (messages, dest) => messages.MoveToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationMailbox_MailboxClosed()
    {
      CopyOrMoveToDestinationMailboxMailboxClosed(new[] {ImapCapability.Move},
                                                  (messages, dest) => messages.MoveTo(dest));
    }

    private void CopyOrMoveToDestinationMailboxMailboxClosed(ImapCapability[] capabilities,
                                                             Action<ImapMessageInfoBase, ImapMailboxInfo> testAction)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("LIST ",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = messageSet.Client.GetMailbox("dest");

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));

        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapUnavailableException>(() => testAction(messageSet, destMailbox));

        Assert.IsFalse(messageSet.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName(server, messageSet, messages, false, (m, name) => m.CopyTo(name));
      });
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName(server, messageSet, messages, false, (m, name) => m.CopyToAsync(name).Wait());
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName(server, messageSet, messages, true, (m, name) => m.MoveTo(name));
      });
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName(server, messageSet, messages, true, (m, name) => m.MoveToAsync(name).Wait());
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_MoveDisabled()
    {
      TestMessageSet(3, new[] {ImapCapability.Move}, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        messageSet.Client.DisabledCapabilities.Add(ImapCapability.Move);

        CopyOrMoveToDestinationName(server, messageSet, messages, true, (m, name) => m.MoveTo(name));
      });
    }

    private void CopyOrMoveToDestinationName(ImapPseudoServer server,
                                             ImapMessageInfoBase messageSet,
                                             ImapMessageInfo[] messages,
                                             bool move,
                                             Action<ImapMessageInfoBase, string> testAction)
    {
      var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                     "$tag OK done\r\n");
        if (move)
          server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
      });

      testAction(messageSet, "dest");

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      if (move) {
        Assert.IsFalse(messageSet.IsAvailable);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_TryCreate()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName_TryCreate(server, messageSet, messages, false, (m, name) => m.CopyTo(name, true));
      });
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_TryCreate()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName_TryCreate(server, messageSet, messages, false, (m, name) => m.CopyToAsync(name, true).Result);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_TryCreate()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName_TryCreate(server, messageSet, messages, true, (m, name) => m.MoveTo(name, true));
      });
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_TryCreate()
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        CopyOrMoveToDestinationName_TryCreate(server, messageSet, messages, true, (m, name) => m.MoveToAsync(name, true).Result);
      });
    }

    private void CopyOrMoveToDestinationName_TryCreate(ImapPseudoServer server,
                                                       ImapMessageInfoBase messageSet,
                                                       ImapMessageInfo[] messages,
                                                       bool move,
                                                       Func<ImapMessageInfoBase, string, ImapMailboxInfo> testFunc)
    {
      var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
        server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                     "$tag OK done\r\n");
        if (move)
          server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                       "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                       "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                       "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                       "$tag OK done\r\n");
      });

      var createdMailbox = testFunc(messageSet, "dest");

      responseTask.Wait();

      Assert.IsFalse(messageSet.Client.IsBusy);

      Assert.IsNull(createdMailbox);

      if (move) {
        Assert.IsFalse(messageSet.IsAvailable);

        Assert.IsNotNull(messages[0].Flags);
        Assert.AreEqual(1, messages[0].Flags.Count);
        Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[0].IsMarkedAsDeleted);

        Assert.IsNotNull(messages[1].Flags);
        Assert.AreEqual(2, messages[1].Flags.Count);
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messages[1].IsMarkedAsDeleted);
        Assert.IsTrue(messages[1].IsSeen);

        Assert.IsNotNull(messages[2].Flags);
        Assert.AreEqual(2, messages[2].Flags.Count);
        Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messages[2].Flags.Contains("$label1"));
        Assert.IsTrue(messages[2].IsMarkedAsDeleted);
      }
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName()
    {
      MoveToMoveCapableDestinationName(false);
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_NoTryCreateResponse()
    {
      MoveToMoveCapableDestinationName(true);
    }

    private void MoveToMoveCapableDestinationName(bool tryCreate)
    {
      TestMessageSet(3,
                     new[] {ImapCapability.Move},
                     ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          if (tryCreate) {
            server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                         "$tag NO [TRYCREATE] mailbox not exist\r\n");
            server.EnqueueTaggedResponse("CREATE dest\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                         "* EXPUNGE 3\r\n" +
                                         "* EXPUNGE 2\r\n" +
                                         "* EXPUNGE 1\r\n" +
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                         "* LIST () \"\" dest\r\n" +
                                         "$tag OK done\r\n");
          }
          else {
            server.EnqueueTaggedResponse("UID MOVE 1:3 dest\r\n",
                                         "* EXPUNGE 3\r\n" +
                                         "* EXPUNGE 2\r\n" +
                                         "* EXPUNGE 1\r\n" +
                                         "$tag OK done\r\n");
          }
        });

        var sourceMailbox = messageSet.Mailbox;
        var createdMailbox = messageSet.MoveTo("dest", tryCreate);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        if (tryCreate) {
          Assert.IsNotNull(createdMailbox);
          Assert.AreEqual("dest", createdMailbox.Name);
          Assert.AreEqual("dest", createdMailbox.FullName);
        }
        else {
          Assert.IsNull(createdMailbox);
        }

        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));
        Assert.AreEqual(0L, sourceMailbox.ExistMessageCount);

        Assert.IsFalse(messages[0].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[0].Sequence);
        Assert.IsTrue(messages[0].IsDeleted);

        Assert.IsFalse(messages[1].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[1].Sequence);
        Assert.IsTrue(messages[1].IsDeleted);

        Assert.IsFalse(messages[2].IsAvailable);
        Assert.AreEqual(ImapMessageInfo.ExpungedMessageSequenceNumber, messages[2].Sequence);
        Assert.IsTrue(messages[2].IsDeleted);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(Array.Empty<ImapCapability>(),
                                             false,
                                             (messages, name) => messages.CopyTo(name));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(Array.Empty<ImapCapability>(),
                                             false,
                                             (messages, name) => messages.CopyToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(Array.Empty<ImapCapability>(),
                                             true,
                                             (messages, name) => messages.MoveTo(name));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(Array.Empty<ImapCapability>(),
                                             true,
                                             (messages, name) => messages.MoveToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_SameMailbox()
    {
      CopyOrMoveToDestinationNameSameMailbox(new[] {ImapCapability.Move},
                                             true,
                                             (messages, name) => messages.MoveTo(name));
    }

    private void CopyOrMoveToDestinationNameSameMailbox(ImapCapability[] capabilities,
                                                        bool move,
                                                        Action<ImapMessageInfoBase, string> testAction)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
        Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);

        if (move) {
          // do nothing
          testAction(messageSet, messageSet.Mailbox.FullName);
        }
        else {
          var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
            server.EnqueueTaggedResponse("UID COPY 1:3 " + messageSet.Mailbox.FullName + "\r\n",
                                         "* 3 RECENT\r\n" +
                                         "* 6 EXISTS\r\n" +
                                         "$tag OK done\r\n");
          });

          testAction(messageSet, messageSet.Mailbox.FullName);

          responseTask.Wait();
        }

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);

        if (move) {
          Assert.AreEqual(3, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(0, messageSet.Mailbox.RecentMessageCount);
        }
        else {
          Assert.AreEqual(6, messageSet.Mailbox.ExistMessageCount);
          Assert.AreEqual(3, messageSet.Mailbox.RecentMessageCount);
        }
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.CopyTo(name));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.CopyToAsync(name));
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.MoveTo(name));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.MoveToAsync(name));
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_ArgumentEmpty()
    {
      CopyOrMoveToDestinationNameArgumentEmpty(new[] {ImapCapability.Move},
                                               (messages, name) => messages.MoveTo(name));
    }

    private void CopyOrMoveToDestinationNameArgumentEmpty(ImapCapability[] capabilities,
                                                          Action<ImapMessageInfoBase, string> testAction)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentException>(() => testAction(messageSet, string.Empty));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(Array.Empty<ImapCapability>(),
                                              (messages, name) => messages.CopyTo(name));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(Array.Empty<ImapCapability>(),
                                              (messages, name) => messages.CopyToAsync(name));
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(Array.Empty<ImapCapability>(),
                                              (messages, name) => messages.MoveTo(name));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(Array.Empty<ImapCapability>(),
                                              (messages, name) => messages.MoveToAsync(name));
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_ArgumentNull()
    {
      CopyOrMoveToDestinationNameArgumentNull(new[] {ImapCapability.Move},
                                              (messages, name) => messages.MoveTo(name));
    }

    private void CopyOrMoveToDestinationNameArgumentNull(ImapCapability[] capabilities,
                                                         Action<ImapMessageInfoBase, string> testAction)
    {
      TestMessageSet(3, capabilities, ImapSequenceSet.CreateUidSet(1L, 3L, 5L),
                     (server, messageSet, messages) => {
        Assert.Throws<ArgumentNullException>(() => testAction(messageSet, (string)null));

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(Array.Empty<ImapCapability>(),
                                          (messages, name) => messages.CopyTo(name));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(Array.Empty<ImapCapability>(),
                                          (messages, name) => messages.CopyToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(Array.Empty<ImapCapability>(),
                                          (messages, name) => messages.MoveTo(name));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(Array.Empty<ImapCapability>(),
                                          (messages, name) => messages.MoveToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_EmptySet()
    {
      CopyOrMoveToDestinationNameEmptySet(new[] {ImapCapability.Move},
                                          (messages, name) => messages.MoveTo(name));
    }

    private void CopyOrMoveToDestinationNameEmptySet(ImapCapability[] capabilities,
                                                     Action<ImapMessageInfoBase, string> testAction)
    {
      TestMessageSet(0, capabilities, CreateEmptySet(),
                     (server, messageSet, messages) => {
        // do nothing
        testAction(messageSet, "dest");

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestCopyTo_CopyToDestinationName_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveToDestinationName_UidValidityChanged()
    {
    }

    [Test, Ignore("not implemented; uid validity")]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_UidValidityChanged()
    {
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.CopyTo(name));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.CopyToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.MoveTo(name));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(Array.Empty<ImapCapability>(),
                                               (messages, name) => messages.MoveToAsync(name));
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_MailboxClosed()
    {
      CopyOrMoveToDestinationNameMailboxClosed(new[] {ImapCapability.Move},
                                               (messages, name) => messages.MoveTo(name));
    }

    private void CopyOrMoveToDestinationNameMailboxClosed(ImapCapability[] capabilities,
                                                          Action<ImapMessageInfoBase, string> testAction)
    {
      TestMessageSet(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     (server, messageSet, messages) => {
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");

        var mailbox = messageSet.Mailbox;

        mailbox.Close();

        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(messageSet.IsAvailable);
        Assert.Throws<ImapUnavailableException>(() => Assert.IsNull(messageSet.Mailbox));

        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapUnavailableException>(() => testAction(messageSet, "dest"));

        Assert.IsFalse(messageSet.Client.IsBusy);

        TestUtils.Assert.IsSerializableBinaryFormat(ex);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(false, Array.Empty<ImapCapability>());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true, Array.Empty<ImapCapability>());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_NoResponse()
    {
      CopyOrMoveToDestinationNameNoResponse(true, new[] {ImapCapability.Move});
    }

    private void CopyOrMoveToDestinationNameNoResponse(bool move, ImapCapability[] capabilities)
    {
      TestMessageSet(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse((move && 0 < capabilities.Length) ? "UID MOVE 1:10 dest\r\n" : "UID COPY 1:10 dest\r\n",
                                       "$tag NO error occurred\r\n");
        });

        if (move)
          Assert.Throws<ImapErrorResponseException>(() => messageSet.MoveTo("dest", false));
        else
          Assert.Throws<ImapErrorResponseException>(() => messageSet.CopyTo("dest", false));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(Array.Empty<ImapCapability>(),
                                                         false,
                                                         (messages, name) => messages.CopyTo(name));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(Array.Empty<ImapCapability>(),
                                                         false,
                                                         (messages, name) => messages.CopyToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(Array.Empty<ImapCapability>(),
                                                         true,
                                                         (messages, name) => messages.MoveTo(name));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(Array.Empty<ImapCapability>(),
                                                         true,
                                                         (messages, name) => messages.MoveToAsync(name).Wait());
    }

    [Test]
    public void TestMoveTo_MoveCapable_MoveToDestinationName_NoResponse_DontTryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseDontTryCreate(new[] {ImapCapability.Move},
                                                         true,
                                                         (messages, name) => messages.MoveTo(name));
    }

    private void CopyOrMoveToDestinationNameNoResponseDontTryCreate(ImapCapability[] capabilities,
                                                                    bool move,
                                                                    Action<ImapMessageInfoBase, string> testAction)
    {
      TestMessageSet(10, capabilities, ImapSequenceSet.CreateUidRangeSet(1L, 10L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse((move && 0 < capabilities.Length) ? "UID MOVE 1:10 dest\r\n" : "UID COPY 1:10 dest\r\n",
                                       "$tag NO [TRYCREATE] mailbox not exist\r\n");
        });

        var ex = TestUtils.Assert.ThrowsOrAggregates<ImapMailboxNotFoundException>(() => testAction(messageSet, "dest"));

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("dest", ex.Mailbox);

        TestUtils.Assert.IsSerializableBinaryFormat(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });


        Assert.IsTrue(messageSet.IsAvailable);
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse_TryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(false,
                                                     (messages, name) => messages.CopyTo(name, true));
    }

    [Test]
    public void TestCopyToAsync_CopyToDestinationName_NoResponse_TryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(false,
                                                     (messages, name) => messages.CopyToAsync(name, true).Result);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse_TryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(true,
                                                     (messages, name) => messages.MoveTo(name, true));
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationName_NoResponse_TryCreate()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreate(true,
                                                     (messages, name) => messages.MoveToAsync(name, true).Result);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreate(bool move,
                                                                Func<ImapMessageInfoBase, string, ImapMailboxInfo> testFunc)
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                       "$tag NO [TRYCREATE] mailbox not exist\r\n");
          server.EnqueueTaggedResponse("CREATE dest\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID COPY 1:3 dest\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                       "* LIST () \"\" dest\r\n" +
                                       "$tag OK done\r\n");
          if (move)
            server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                         "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                         "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                         "$tag OK done\r\n");
        });

        var createdMailbox = testFunc(messageSet, "dest");

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("dest", createdMailbox.Name);
        Assert.AreEqual("dest", createdMailbox.FullName);

        if (move) {
          Assert.IsFalse(messageSet.IsAvailable);

          Assert.IsNotNull(messages[0].Flags);
          Assert.AreEqual(1, messages[0].Flags.Count);
          Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[0].IsMarkedAsDeleted);

          Assert.IsNotNull(messages[1].Flags);
          Assert.AreEqual(2, messages[1].Flags.Count);
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(messages[1].IsMarkedAsDeleted);
          Assert.IsTrue(messages[1].IsSeen);

          Assert.IsNotNull(messages[2].Flags);
          Assert.AreEqual(2, messages[2].Flags.Count);
          Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[2].Flags.Contains("$label1"));
          Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        }
      });
    }

    [Test]
    public void TestCopyTo_CopyToDestinationName_NoResponse_TryCreate_NameContainsWildcard()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(true);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationName_NoResponse_TryCreate_NameContainsWildcard()
    {
      CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(false);
    }

    private void CopyOrMoveToDestinationNameNoResponseTryCreateNameContainsWildcard(bool move)
    {
      TestMessageSet(3, ImapSequenceSet.CreateUidRangeSet(1L, 3L),
                     (server, messageSet, messages) => {
        var responseTask = TestCases.QueueResponseTask(server, messageSet.Client, delegate {
          server.EnqueueTaggedResponse("UID COPY 1:3 \"Box*\"\r\n",
                                       "$tag NO [TRYCREATE] mailbox not exist\r\n");
          server.EnqueueTaggedResponse("CREATE \"Box*\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UID COPY 1:3 \"Box*\"\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Box*\r\n",
                                       "* LIST (\\Marked) \".\" Box*\r\n" +
                                       "* LIST () \".\" Box\r\n" +
                                       "* LIST () \".\" Box.Child\r\n" +
                                       "* LIST () \".\" Boxes\r\n" +
                                       "$tag OK done\r\n");
          if (move)
            server.EnqueueTaggedResponse("UID STORE 1:3 +FLAGS (\\Deleted)\r\n",
                                         "* FETCH 1 (FLAGS (\\Deleted))\r\n" +
                                         "* FETCH 2 (FLAGS (\\Deleted \\Seen))\r\n" +
                                         "* FETCH 3 (FLAGS (\\Deleted $label1))\r\n" +
                                         "$tag OK done\r\n");
        });

        var createdMailbox = move
          ? messageSet.MoveTo("Box*", true)
          : messageSet.CopyTo("Box*", true);

        responseTask.Wait();

        Assert.IsFalse(messageSet.Client.IsBusy);

        Assert.IsNotNull(createdMailbox);
        Assert.AreEqual("Box*", createdMailbox.Name);
        Assert.AreEqual("Box*", createdMailbox.FullName);
        Assert.AreEqual(1, createdMailbox.Flags.Count);
        CollectionAssert.Contains(createdMailbox.Flags, ImapMailboxFlag.Marked);

        if (move) {
          Assert.IsFalse(messageSet.IsAvailable);

          Assert.IsNotNull(messages[0].Flags);
          Assert.AreEqual(1, messages[0].Flags.Count);
          Assert.IsTrue(messages[0].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[0].IsMarkedAsDeleted);

          Assert.IsNotNull(messages[1].Flags);
          Assert.AreEqual(2, messages[1].Flags.Count);
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[1].Flags.Contains(ImapMessageFlag.Seen));
          Assert.IsTrue(messages[1].IsMarkedAsDeleted);
          Assert.IsTrue(messages[1].IsSeen);

          Assert.IsNotNull(messages[2].Flags);
          Assert.AreEqual(2, messages[2].Flags.Count);
          Assert.IsTrue(messages[2].Flags.Contains(ImapMessageFlag.Deleted));
          Assert.IsTrue(messages[2].Flags.Contains("$label1"));
          Assert.IsTrue(messages[2].IsMarkedAsDeleted);
        }
      });
    }
  }
}
