using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Text.Encodings;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessagePartInfoOperationsFetchTests {
    [SetUp]
    public void SetUp()
    {
      TestUtils.Encodings.RegisterCodePagesEncodingProvider();
    }

    private const string dummyEnvelope = "(NIL \"test\" NIL NIL NIL NIL NIL NIL NIL NIL)";

    private void TestMessage(string fetchResponseBodyStructure,
                             Action<ImapPseudoServer, ImapMessageInfo> testAction)
    {
      var fetchResp =
        "* FETCH 1 (UID 1 " +
        "BODYSTRUCTURE " + fetchResponseBodyStructure +
        " ENVELOPE " + dummyEnvelope +
        " INTERNALDATE \"27-Jul-2014 21:50:32 +0900\" RFC822.SIZE 42)\r\n" +
        "$tag OK done\r\n";

      TestCases.WithMessage(fetchResp,
                             ImapMessageFetchAttributeOptions.StaticAttributes,
                             (server, message) => {
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.MessagePart);

        testAction(server, message);
      });
    }

    private void TestMessagePartPlainTextSingle(Action<ImapPseudoServer, ImapMessagePartInfo> testAction)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        testAction(server, part);
      });
    }

    private void TestMessagePartPlainTextSingle(string messageBody, Action<ImapPseudoServer, ImapMessagePartInfo, string> testAction)
    {
      messageBody = messageBody.Replace("\r\n", "\n").Replace("\n", "\r\n");
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsNotNull(part);
        Assert.AreEqual(part.Length, messageOctet);

        testAction(server, part, messageBody);
      });
    }

    [Test]
    public void TestOpenRead_SinglePart()
    {
      OpenRead_SinglePart(part => part.OpenRead());
    }

    [Test]
    public void TestOpenReadAsync_SinglePart()
    {
      OpenRead_SinglePart(part => part.OpenReadAsync().Result);
    }

    private void OpenRead_SinglePart(Func<ImapMessagePartInfo, Stream> openRead)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        Assert.IsFalse(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreEqual(expectedMessageBody, (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());
        Assert.IsFalse(part.Message.IsSeen);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_Encapsulated()
    {
      OpenRead_Encapsulated(part => part.OpenRead());
    }

    [Test]
    public void TestOpenReadAsync_Encapsulated()
    {
      OpenRead_Encapsulated(part => part.OpenReadAsync().Result);
    }

    private void OpenRead_Encapsulated(Func<ImapMessagePartInfo, Stream> openRead)
    {
      const string bodyStructureEncapsulated = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "(" +
        "(\"message\" \"rfc822\" NIL NIL NIL \"7bit\" 160 " +
        dummyEnvelope + " " + 
        bodyStructureEncapsulated + " 1)" +
        " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsFalse(part.IsMultiPart);
        Assert.IsTrue(part.HasEncapsulatedMessage);

        const string expectedMessageBody = @"MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: message/rfc822

MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: text/plain

0123456789abcdef
";

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreEqual(expectedMessageBody,
                        (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());
        Assert.IsFalse(part.Message.IsSeen);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_MultiPart()
    {
      OpenRead_MultiPart(part => part.OpenRead());
    }

    [Test]
    public void TestOpenReadAsync_MultiPart()
    {
      OpenRead_MultiPart(part => part.OpenReadAsync().Result);
    }

    private void OpenRead_MultiPart(Func<ImapMessagePartInfo, Stream> openRead)
    {
      const string bodyStructureNested = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "((" + bodyStructureNested + " \"mixed\") \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsTrue(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        const string expectedMessageBody = @"MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: multipart/mixed; boundary=""xxxxxxxx""

This is a multi-part message in MIME format.

--xxxxxxxx
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: text/plain

0123456789abcdef
--xxxxxxxx--
";

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreEqual(expectedMessageBody,
                        (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());
        Assert.IsFalse(part.Message.IsSeen);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_OmitHeader_RootMessagePart()
    {
      OpenRead_OmitHeader_RootMessagePart(part => part.OpenRead(ImapMessageFetchBodyOptions.OmitHeader));
    }

    [Test]
    public void TestOpenReadAsync_OmitHeader_RootMessagePart()
    {
      OpenRead_OmitHeader_RootMessagePart(part => part.OpenReadAsync(ImapMessageFetchBodyOptions.OmitHeader).Result);
    }

    private void OpenRead_OmitHeader_RootMessagePart(Func<ImapMessagePartInfo, Stream> openRead)
    {
      const string respBodyStructure = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string body = "0123456789abcdef";

      TestMessage(respBodyStructure, (server, message) => {
        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[TEXT]<0.10240>)\r\n",
                                       "* 1 FETCH (BODY[1] {16}\r\n" + body + ")\r\n" +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(message.MessagePart);

        responseTask.Wait();

        Assert.IsTrue(message.Client.IsBusy);

        Assert.IsNotNull(stream);
        Assert.AreEqual(0L, stream.Length);

        stream.Dispose();

        Assert.IsFalse(message.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_OmitHeader_SpecificMessagePart()
    {
      OpenRead_OmitHeader_SpecificMessagePart(part => part.OpenRead(ImapMessageFetchBodyOptions.OmitHeader));
    }

    [Test]
    public void TestOpenReadAsync_OmitHeader_SpecificMessagePart()
    {
      OpenRead_OmitHeader_SpecificMessagePart(part => part.OpenReadAsync(ImapMessageFetchBodyOptions.OmitHeader).Result);
    }

    private void OpenRead_OmitHeader_SpecificMessagePart(Func<ImapMessagePartInfo, Stream> openRead)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       "* 1 FETCH (BODY[1] {16}\r\n" + body + ")\r\n" +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        Assert.AreEqual(0L, stream.Length);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_RangedFetch()
    {
      OpenRead_RangedFetch((part, range) => part.OpenRead(range));
    }

#if false
    [Test]
    public void TestOpenReadAsync_RangedFetch()
    {
      OpenRead_RangedFetch((part, range) => part.OpenReadAsync(range).Result);
    }
#endif

    private void OpenRead_RangedFetch(Func<ImapMessagePartInfo, ImapPartialRange, Stream> openRead)
    {
      var fetchRange = new ImapPartialRange(128, 64);
      var body = new string('x', (int)fetchRange.Length);

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {

        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<128.64>)\r\n",
                                       "* 1 FETCH (BODY[1] {64}\r\n" + body + ")\r\n" +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part, fetchRange);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        Assert.AreEqual(fetchRange.Length,stream.Length);
        Assert.AreEqual(expectedMessageBody, (new StreamReader(stream, OctetEncoding.EightBits)).ReadToEnd());
        Assert.IsFalse(part.Message.IsSeen);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_RangedFetch_DecodeContent()
    {
      OpenRead_RangedFetch_DecodeContent((part, range) => part.OpenRead(range, ImapMessageFetchBodyOptions.DecodeContent));
    }

#if false
    [Test]
    public void TestOpenReadAsync_RangedFetch()
    {
    OpenRead_RangedFetch_DecodeContent((part, range) => part.OpenReadAsync(range, ImapMessageFetchBodyOptions.DecodeContent).Result);
    }
#endif

    private void OpenRead_RangedFetch_DecodeContent(Func<ImapMessagePartInfo, ImapPartialRange, Stream> openRead)
    {
      var fetchRange = new ImapPartialRange(128, 64);
      var body = new string('x', (int)fetchRange.Length);

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        TestUtils.Assert.ThrowsOrAggregates<NotSupportedException>(() => openRead(part, fetchRange));

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_DecodeContent_CTEBase64()
    {
      OpenRead_DecodeContent_CTEBase64(part => part.OpenRead(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenReadAsync_DecodeContent_CTEBase64()
    {
      OpenRead_DecodeContent_CTEBase64(part => part.OpenReadAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenRead_DecodeContent_CTEBase64(Func<ImapMessagePartInfo, Stream> openRead)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreEqual("base64 encoded 本文",
                        (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());
        Assert.IsFalse(part.Message.IsSeen);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_DecodeContent_CTEQuotedPrintable()
    {
      OpenRead_DecodeContent_CTEQuotedPrintable(part => part.OpenRead(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenReadAsync_DecodeContent_CTEQuotedPrintable()
    {
      OpenRead_DecodeContent_CTEQuotedPrintable(part => part.OpenReadAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenRead_DecodeContent_CTEQuotedPrintable(Func<ImapMessagePartInfo, Stream> openRead)
    {
      var messageBody = @"quoted-printable =E6=9C=AC=E6=96=87";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"quoted-printable\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var stream = openRead(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(stream);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreEqual("quoted-printable 本文",
                        (new StreamReader(stream, Encoding.UTF8)).ReadToEnd());
        Assert.IsFalse(part.Message.IsSeen);

        stream.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_DecodeContent_CloseStreamBeforeRead_Base64()
    {
      OpenRead_DecodeContent_CloseStreamBeforeRead_Base64(part => part.OpenRead(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenReadAsync_DecodeContent_CloseStreamBeforeRead_Base64()
    {
      OpenRead_DecodeContent_CloseStreamBeforeRead_Base64(part => part.OpenReadAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenRead_DecodeContent_CloseStreamBeforeRead_Base64(Func<ImapMessagePartInfo, Stream> openRead)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });


        var stream = openRead(part);

        responseTask.Wait();

        Assert.DoesNotThrow(() => stream.Dispose());

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenRead_DecodeContent_CloseStreamBeforeRead_QuotedPrintable()
    {
      OpenRead_DecodeContent_CloseStreamBeforeRead_QuotedPrintable(part => part.OpenRead(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenReadAsync_DecodeContent_CloseStreamBeforeRead_QuotedPrintable()
    {
      OpenRead_DecodeContent_CloseStreamBeforeRead_QuotedPrintable(part => part.OpenReadAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenRead_DecodeContent_CloseStreamBeforeRead_QuotedPrintable(Func<ImapMessagePartInfo, Stream> openRead)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"application\" \"octet-stream\" NIL NIL NIL \"quoted-printable\" {0})", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });


        var stream = openRead(part);

        responseTask.Wait();

        Assert.DoesNotThrow(() => stream.Dispose());

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream()
    {
      ReadAs_Stream((part, converter) => part.ReadAs(converter));
    }

    [Test]
    public void TestReadAsAsync_Stream()
    {
      ReadAs_Stream((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_Stream(Func<ImapMessagePartInfo, Converter<Stream, byte[]>, byte[]> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAs(part, delegate(Stream stream) {
          Assert.IsNotNull(stream);
          //Assert.AreEqual(part.Length, stream.Length);

          Assert.IsTrue(part.Client.IsBusy);

          return stream.ReadToEnd();
        });

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody,
                        OctetEncoding.EightBits.GetString(ret));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_ExceptionInConveter()
    {
      ReadAs_Stream_ExceptionInConveter((part, converter) => part.ReadAs(converter));
    }

    [Test]
    public void TestReadAsAsync_Stream_ExceptionInConveter()
    {
      ReadAs_Stream_ExceptionInConveter((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_Stream_ExceptionInConveter(Func<ImapMessagePartInfo, Converter<Stream, int>, int> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part,  expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Assert.ThrowsOrAggregates<NotSupportedException>(() => readAs(part, (Converter<Stream, int>)delegate {
          throw new NotSupportedException("exception");
        }));

        responseTask.Wait();

        Assert.AreEqual("exception", ex.Message);

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_Stream_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAs<string>((Converter<Stream, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAsAsync_Stream_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAsAsync<string>((Converter<Stream, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAllBytes()
    {
      ReadAllBytes(part => part.ReadAllBytes());
    }

    [Test]
    public void TestReadAllBytesAsync()
    {
      ReadAllBytes(part => part.ReadAllBytesAsync().Result);
    }

    private void ReadAllBytes(Func<ImapMessagePartInfo, byte[]> readAllBytes)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part,expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAllBytes(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody,
                        OctetEncoding.EightBits.GetString(ret));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteTo_Stream()
    {
      WriteTo_Stream((part, dest) => part.WriteTo(dest));
    }

    [Test]
    public void TestWriteToAsync_Stream()
    {
      WriteTo_Stream((part, dest) => part.WriteToAsync(dest).Wait());
    }

    private void WriteTo_Stream(Action<ImapMessagePartInfo, Stream> writeTo)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        using (var outputStream = new MemoryStream()) {
          writeTo(part, outputStream);

          responseTask.Wait();

          outputStream.Dispose();

          //Assert.AreEqual(part.Length, outputStream.Length);
          Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        outputStream.ToArray());
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }

    [Test]
    public void TestWriteTo_Stream_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteTo((Stream)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteTo_Stream_DestinationNotWritable()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => part.WriteTo(stream));
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }

    [Test]
    public void TestWriteToAsync_Stream_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteToAsync((Stream)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteToAsync_Stream_DestinationNotWritable()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => part.WriteToAsync(stream));
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }

    [Test]
    public void TestWriteTo_BinaryWriter()
    {
      WriteTo_BinaryWriter((part, writer) => part.WriteTo(writer));
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter()
    {
      WriteTo_BinaryWriter((part, writer) => part.WriteToAsync(writer).Wait());
    }

    private void WriteTo_BinaryWriter(Action<ImapMessagePartInfo, BinaryWriter> writeTo)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body,
                                     (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        using (var outputStream = new MemoryStream()) {
          var writer = new BinaryWriter(outputStream);

          writeTo(part, writer);

          responseTask.Wait();

          outputStream.Dispose();

          //Assert.AreEqual(part.Length, outputStream.Length);
          Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                          outputStream.ToArray());
          Assert.IsFalse(part.Client.IsBusy);
        }
      });
    }


    [Test]
    public void TestWriteTo_BinaryWriter_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteTo((BinaryWriter)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter_DestinationNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.WriteToAsync((BinaryWriter)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestSave()
    {
      Save((part, file) => part.Save(file));
    }

    [Test]
    public void TestSaveAsync()
    {
      Save((part, file) => part.SaveAsync(file).Wait());
    }

    private void Save(Action<ImapMessagePartInfo, string> save)
    {
      const string body = "0123456789abcdef";
      var outfile = Path.Combine(TestContext.CurrentContext.WorkDirectory, "fetched.eml");
      const int existingFileLength = 10 * 1024;

      TestUtils.UsingFile(outfile, () => {
        // prepare existing file
        File.WriteAllText(outfile, new string('x', existingFileLength));

        TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
          var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
            server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
          });

          save(part, outfile);

          responseTask.Wait();

          Assert.IsTrue(File.Exists(outfile), "file existence");
          //Assert.AreEqual(part.Length, (new FileInfo(outfile)).Length);
          Assert.AreEqual(expectedMessageBody, File.ReadAllText(outfile));
          Assert.IsFalse(part.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestSave_PathNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.Save(null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestSave_PathEmpty()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentException>(() => part.Save(string.Empty));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestSaveAsync_PathNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.SaveAsync(null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestSaveAsync_PathEmpty()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentException>(() => part.SaveAsync(string.Empty));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestDownload()
    {
      Download(part => part.Download());
    }

    [Test]
    public void TestDownloadAsync()
    {
      Download(part => part.DownloadAsync().Result);
    }

    private void Download(Func<ImapMessagePartInfo, Stream> download)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = download(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.IsTrue(ret.CanSeek, "CanSeek");
        Assert.IsTrue(ret.CanRead, "CanRead");
        Assert.IsTrue(ret.CanWrite, "CanWrite");
        Assert.AreEqual(0L, ret.Position, "Position");
        //Assert.AreEqual(part.Length, ret.Length, "Length");

        Assert.DoesNotThrow(() => ret.Seek(-1L, SeekOrigin.End));
        Assert.AreNotEqual(-1, ret.ReadByte());
        Assert.DoesNotThrow(() => ret.Seek(0L, SeekOrigin.Begin));

        Assert.AreEqual(OctetEncoding.EightBits.GetBytes(expectedMessageBody),
                        ret.ReadToEnd());

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_NotDecodeContent()
    {
      OpenText_NotDecodeContent(part => part.OpenText());
    }

    [Test]
    public void TestOpenTextAsync_NotDecodeContent()
    {
      OpenText_NotDecodeContent(part => part.OpenTextAsync().Result);
    }

    private void OpenText_NotDecodeContent(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"0123456789abcdef";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"7bit\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(reader);
        //Assert.AreEqual(part.Length, reader.BaseStream.Length);
        Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
        Assert.AreEqual(messageBody, reader.ReadToEnd());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_NotDecodeContent_CharsetMissing()
    {
      OpenText_NotDecodeContent_CharsetMissing(part => part.OpenText());
    }

    [Test]
    public void TestOpenTextAsync_NotDecodeContent_CharsetMissing()
    {
      OpenText_NotDecodeContent_CharsetMissing(part => part.OpenTextAsync().Result);
    }

    private void OpenText_NotDecodeContent_CharsetMissing(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"0123456789abcdef";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" NIL NIL NIL \"7bit\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(reader);
        //Assert.AreEqual(part.Length, reader.BaseStream.Length);
        Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
        Assert.AreEqual(messageBody, reader.ReadToEnd());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_CTEBase64()
    {
      OpenText_DecodeContent_CTEBase64(part => part.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_CTEBase64()
    {
      OpenText_DecodeContent_CTEBase64(part => part.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenText_DecodeContent_CTEBase64(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(reader);
        //Assert.AreEqual(part.Length, reader.BaseStream.Length);
        Assert.AreEqual(Encoding.UTF8, reader.CurrentEncoding);
        Assert.AreEqual("base64 encoded 本文", reader.ReadToEnd());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_CTEQuotedPrintable()
    {
      OpenText_DecodeContent_CTEQuotedPrintable(part => part.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_CTEQuotedPrintable()
    {
      OpenText_DecodeContent_CTEQuotedPrintable(part => part.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenText_DecodeContent_CTEQuotedPrintable(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"quoted-printable =E6=9C=AC=E6=96=87";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"quoted-printable\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(reader);
        //Assert.AreEqual(part.Length, reader.BaseStream.Length);
        Assert.AreEqual(Encoding.UTF8, reader.CurrentEncoding);
        Assert.AreEqual("quoted-printable 本文", reader.ReadToEnd());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_CloseReaderBeforeRead_Base64()
    {
      OpenText_DecodeContent_CloseReaderBeforeRead_Base64(part => part.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_CloseReaderBeforeRead_Base64()
    {
      OpenText_DecodeContent_CloseReaderBeforeRead_Base64(part => part.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenText_DecodeContent_CloseReaderBeforeRead_Base64(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.DoesNotThrow(() => reader.Dispose());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_CloseReaderBeforeRead_QuotedPrintable()
    {
      OpenText_DecodeContent_CloseReaderBeforeRead_QuotedPrintable(part => part.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_CloseReaderBeforeRead_QuotedPrintable()
    {
      OpenText_DecodeContent_CloseReaderBeforeRead_QuotedPrintable(part => part.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenText_DecodeContent_CloseReaderBeforeRead_QuotedPrintable(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"========";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"utf-8\") NIL NIL \"quoted-printable\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.DoesNotThrow(() => reader.Dispose());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_CharsetMissing()
    {
      OpenText_DecodeContent_CharsetMissing(part => part.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_CharsetMissing()
    {
      OpenText_DecodeContent_CharsetMissing(part => part.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenText_DecodeContent_CharsetMissing(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" NIL NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var reader = openText(part);

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(reader);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreEqual(TestUtils.Encodings.Latin1, reader.CurrentEncoding);
        StringAssert.StartsWith("base64", reader.ReadToEnd());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_CharsetUnsupported()
    {
      OpenText_DecodeContent_CharsetUnsupported(part => part.OpenText(ImapMessageFetchBodyOptions.DecodeContent));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_CharsetUnsupported()
    {
      OpenText_DecodeContent_CharsetUnsupported(part => part.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent).Result);
    }

    private void OpenText_DecodeContent_CharsetUnsupported(Func<ImapMessagePartInfo, StreamReader> openText)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"x-unknown-charset\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.Throws<EncodingNotSupportedException>(() => openText(part));

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_FallbackCharsetAvailable()
    {
      OpenText_DecodeContent_FallbackCharsetAvailable((message, callback) => message.OpenText(ImapMessageFetchBodyOptions.DecodeContent, callback));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_FallbackCharsetAvailable()
    {
      OpenText_DecodeContent_FallbackCharsetAvailable((message, callback) => message.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent, callback).Result);
    }

    private void OpenText_DecodeContent_FallbackCharsetAvailable(Func<ImapMessagePartInfo, EncodingSelectionCallback, StreamReader> openText)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"x-unknown-charset\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     messageOctet,
                                                     messageBody) +
                                       "$tag OK done\r\n");
        });

        var called = false;
        var fallbackCharset = Encoding.UTF8;
        var callback = (EncodingSelectionCallback)delegate(string charset) {
          called = true;

          Assert.AreEqual("x-unknown-charset", charset);

          return fallbackCharset;
        };

        var reader = openText(part, callback);

        Assert.IsTrue(called, "called back");

        responseTask.Wait();

        Assert.IsTrue(part.Client.IsBusy);

        Assert.IsNotNull(reader);
        //Assert.AreEqual(part.Length, stream.Length);
        Assert.AreSame(fallbackCharset, reader.CurrentEncoding);
        StringAssert.StartsWith("base64", reader.ReadToEnd());

        reader.Dispose();

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpenText_DecodeContent_FallbackCharsetNotAvailable()
    {
      OpenText_DecodeContent_FallbackCharsetNotAvailable((message, callback) => message.OpenText(ImapMessageFetchBodyOptions.DecodeContent, callback));
    }

    [Test]
    public void TestOpenTextAsync_DecodeContent_FallbackCharsetNotAvailable()
    {
      OpenText_DecodeContent_FallbackCharsetNotAvailable((message, callback) => message.OpenTextAsync(ImapMessageFetchBodyOptions.DecodeContent, callback).Result);
    }

    private void OpenText_DecodeContent_FallbackCharsetNotAvailable(Func<ImapMessagePartInfo, EncodingSelectionCallback, StreamReader> openText)
    {
      var messageBody = @"YmFzZTY0IGVuY29kZWQg5pys5paH";
      var messageOctet = OctetEncoding.SevenBits.GetByteCount(messageBody);
      var bodyStructurePart1 = string.Format("(\"text\" \"plain\" (\"charset\" \"x-unknown-charset\") NIL NIL \"base64\" {0} 1)", messageOctet);
      var respBodyStructure =  "(" + bodyStructurePart1 + " \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);
        var called = false;
        var callback = (EncodingSelectionCallback)delegate(string charset) {
          called = true;

          Assert.AreEqual("x-unknown-charset", charset);

          return null;
        };

        Assert.Throws<EncodingNotSupportedException>(() => openText(part, callback));

        Assert.IsTrue(called, "called back");

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_StreamReader()
    {
      ReadAs_StreamReader((part, converter) => part.ReadAs(converter));
    }

    [Test]
    public void TestReadAsAsync_StreamReader()
    {
      ReadAs_StreamReader((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_StreamReader(Func<ImapMessagePartInfo, Converter<StreamReader, string>, string> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAs(part, delegate(StreamReader reader) {
          Assert.IsNotNull(reader);
          //Assert.AreEqual(part.Length, reader.BaseStream.Length);

          Assert.IsTrue(part.Client.IsBusy);

          return reader.ReadToEnd();
        });

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody, ret);
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_StreamReader_ExceptionInConveter()
    {
      ReadAs_StreamReader_ExceptionInConveter((part, converter) => part.ReadAs(converter));
    }

    [Test]
    public void TestReadAsAsync_StreamReader_ExceptionInConveter()
    {
      ReadAs_StreamReader_ExceptionInConveter((part, converter) => part.ReadAsAsync(converter).Result);
    }

    private void ReadAs_StreamReader_ExceptionInConveter(Func<ImapMessagePartInfo, Converter<StreamReader, int>, int> readAs)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Assert.ThrowsOrAggregates<NotSupportedException>(() => readAs(part, (Converter<StreamReader, int>)delegate {
          throw new NotSupportedException("exception");
        }));

        responseTask.Wait();

        Assert.AreEqual("exception", ex.Message);

        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAs_StreamReader_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAs<string>((Converter<StreamReader, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAsAsync_StreamReader_ConverterNull()
    {
      TestMessagePartPlainTextSingle((server, part) => {
        Assert.Throws<ArgumentNullException>(() => part.ReadAsAsync<string>((Converter<StreamReader, string>)null));
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadLines()
    {
      const string body = @"1st line
2nd line
3rd line";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                     string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                   part.Length,
                                                   expectedMessageBody) +
                                     "$tag OK done\r\n");

        var lines = part.ReadLines();

        Assert.IsNotNull(lines);

        Assert.IsTrue(part.Client.IsBusy, "IsBusy; still enumerating");

        CollectionAssert.AreEqual((new StringReader(body)).ReadAllLines(),
                                  lines.ToArray());

        Assert.IsFalse(part.Client.IsBusy, "IsBusy; after enumeration");
      });
    }

    [Test]
    public void TestReadAllLines()
    {
      ReadAllLines(part => part.ReadAllLines());
    }

    [Test]
    public void TestReadAllLinesAsync()
    {
      ReadAllLines(part => part.ReadAllLinesAsync().Result);
    }

    private void ReadAllLines(Func<ImapMessagePartInfo, string[]> readAllLines)
    {
      const string body = @"1st line
2nd line
3rd line";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAllLines(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        CollectionAssert.AreEqual(expectedMessageBody.Split(new[] {"\r\n"}, StringSplitOptions.None), ret);
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestReadAllText()
    {
      ReadAllText(part => part.ReadAllText());
    }

    [Test]
    public void TestReadAllTextAsync()
    {
      ReadAllText(part => part.ReadAllTextAsync().Result);
    }

    private void ReadAllText(Func<ImapMessagePartInfo, string> readAllText)
    {
      const string body = "0123456789abcdef";

      TestMessagePartPlainTextSingle(body, (server, part, expectedMessageBody) => {
        var responseTask = TestCases.QueueResponseTask(server, part.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1] {{{0}}}\r\n{1})\r\n",
                                                     part.Length,
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var ret = readAllText(part);

        responseTask.Wait();

        Assert.IsNotNull(ret);
        Assert.AreEqual(expectedMessageBody, ret);
        Assert.IsFalse(part.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetRawHeader()
    {
      GetRawHeader(part => part.GetRawHeader());
    }

    [Test]
    public void TestGetRawHeaderAsync()
    {
      GetRawHeader(part => part.GetRawHeaderAsync().Result);
    }

    private void GetRawHeader(Func<ImapMessagePartInfo, string> getRawHeader)
    {
      const string bodyStructureNested = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "((" + bodyStructureNested + " \"mixed\") \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsTrue(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        var expectedMessageBody = @"MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Type: text/plain
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1.MIME]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1.MIME] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var headers = getRawHeader(part);

        responseTask.Wait();

        Assert.IsFalse(part.Client.IsBusy);
        Assert.IsFalse(part.Message.IsSeen);

        Assert.AreEqual(expectedMessageBody, headers);
      });
    }

    [Test]
    public void TestGetHeader()
    {
      GetHeader(false, part => part.GetHeader());
    }

    [Test]
    public void TestGetHeader_Decode()
    {
      GetHeader(true, part => part.GetHeader(true));
    }

    [Test]
    public void TestGetHeaderAsync()
    {
      GetHeader(false, part => part.GetHeaderAsync().Result);
    }

    [Test]
    public void TestGetHeaderAsync_Decode()
    {
      GetHeader(true, part => part.GetHeaderAsync(true).Result);
    }

    private void GetHeader(bool decode, Func<ImapMessagePartInfo, KeyValuePair<string, string>[]> getHeader)
    {
      const string bodyStructureNested = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 16 1)";
      const string respBodyStructure = "((" + bodyStructureNested + " \"mixed\") \"mixed\")";

      TestMessage(respBodyStructure, (server, message) => {
        var part = message.GetMessagePart(1);

        Assert.IsTrue(part.IsMultiPart);
        Assert.IsFalse(part.HasEncapsulatedMessage);

        var expectedMessageBody = @"Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        var responseTask = TestCases.QueueResponseTask(server, message.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 1 (BODY.PEEK[1.MIME]<0.10240>)\r\n",
                                       string.Format("* 1 FETCH (BODY[1.MIME] {{{0}}}\r\n{1})\r\n",
                                                     OctetEncoding.SevenBits.GetByteCount(expectedMessageBody),
                                                     expectedMessageBody) +
                                       "$tag OK done\r\n");
        });

        var headers = getHeader(part);

        responseTask.Wait();

        Assert.IsFalse(part.Client.IsBusy);
        Assert.IsFalse(part.Message.IsSeen);

        Assert.IsNotEmpty(headers);
        Assert.AreEqual(3, headers.Length);

        if (decode)
          Assert.AreEqual("テスト", headers.Where(pair => pair.Key == "Subject").First().Value);
        else
          Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", headers.Where(pair => pair.Key == "Subject").First().Value);

        Assert.AreEqual("text/plain; charset=ISO-2022-JP", headers.Where(pair => pair.Key == "Content-Type").First().Value);
      });
    }
  }
}

