using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapFetchDataItemTests {
    [Test]
    public void TestMacro()
    {
      Assert.AreEqual("ALL", ImapFetchDataItem.All.ToString());
    }

    [Test]
    public void TestNonMacro()
    {
      Assert.AreEqual("(UID)", ImapFetchDataItem.Uid.ToString());
    }

    [Test]
    public void TestCombine()
    {
      var fetch1 = ImapFetchDataItem.Envelope;
      var fetch2 = ImapFetchDataItem.Uid;

      Assert.AreEqual("(ENVELOPE UID)", ImapFetchDataItem.Combine(fetch1, fetch2).ToString());
      Assert.AreEqual("(ENVELOPE UID)", (fetch1 + fetch2).ToString());
      Assert.AreEqual("(ENVELOPE UID)", fetch1.CombineWith(fetch2).ToString());

      Assert.AreEqual("(UID ENVELOPE)", (fetch2 + fetch1).ToString());
    }

    [Test]
    public void TestCombine_WithMacro()
    {
      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE UID)", (ImapFetchDataItem.Fast + ImapFetchDataItem.Uid).ToString());
      Assert.AreEqual("(UID FLAGS INTERNALDATE RFC822.SIZE)", (ImapFetchDataItem.Uid + ImapFetchDataItem.Fast).ToString());

      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE ENVELOPE UID)", (ImapFetchDataItem.All + ImapFetchDataItem.Uid).ToString());
      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE ENVELOPE BODY UID)", (ImapFetchDataItem.Full + ImapFetchDataItem.Uid).ToString());
    }

    [Test]
    public void TestBody()
    {
      Assert.AreEqual("(BODY[])", ImapFetchDataItem.BodyText().ToString());
      Assert.AreEqual("(BODY.PEEK[])", ImapFetchDataItem.BodyPeek().ToString());
      Assert.AreEqual("(BODY[])", ImapFetchDataItem.BodyText(false).ToString());
      Assert.AreEqual("(BODY.PEEK[])", ImapFetchDataItem.BodyText(true).ToString());
      Assert.AreEqual("(BODY[heADEr])", ImapFetchDataItem.BodyText("heADEr").ToString());
      Assert.AreEqual("(BODY.PEEK[]<1000.2000>)", ImapFetchDataItem.BodyPeek(1000L, 2000L).ToString());
      Assert.AreEqual("(BODY.PEEK[]<1000.2000>)", ImapFetchDataItem.BodyPeek(new ImapPartialRange(1000L, 2000L)).ToString());
      Assert.AreEqual("(BODY[HEADER (subject date)]<1000.2000>)", ImapFetchDataItem.BodyText("HEADER (subject date)", 1000, 2000).ToString());
      Assert.AreEqual("(BODY.PEEK[HEADER (from to)]<0.1000>)", ImapFetchDataItem.BodyPeek("HEADER (from to)", 0, 1000).ToString());
    }

    [Test]
    public void TestBody_InvalidArgument()
    {
      Assert.Throws<ArgumentException>(() => ImapFetchDataItem.BodyPeek(new ImapPartialRange(1000L)), "#1");
      Assert.Throws<ArgumentException>(() => ImapFetchDataItem.BodyText(new ImapPartialRange(1000L)), "#2");
    }

    [Test]
    public void TestBody_ArgumentOutOfRange()
    {
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.BodyPeek(1000L, 0L), "#1");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.BodyText(1000L, 0L), "#2");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.BodyPeek(-1L, 1L), "#3");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.BodyText(-1L, 1L), "#4");
    }

    [Test]
    public void TestBinary()
    {
      Assert.AreEqual("(BINARY[1.1])", ImapFetchDataItem.Binary("1.1").ToString());
      Assert.AreEqual("(BINARY.PEEK[1.1])", ImapFetchDataItem.BinaryPeek("1.1").ToString());
      Assert.AreEqual("(BINARY[1.1])", ImapFetchDataItem.Binary(false, "1.1").ToString());
      Assert.AreEqual("(BINARY.PEEK[1.1])", ImapFetchDataItem.Binary(true, "1.1").ToString());
      Assert.AreEqual("(BINARY[heADEr])", ImapFetchDataItem.Binary("heADEr").ToString());
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<0.1024>)", ImapFetchDataItem.BinaryPeek("1.1.TEXT", 0L, 1024L).ToString());
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<0.1024>)", ImapFetchDataItem.BinaryPeek("1.1.TEXT", new ImapPartialRange(0L, 1024L)).ToString());
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<1000.2000>)", ImapFetchDataItem.BinaryPeek("1.1.TEXT", 1000L, 2000L).ToString());
      Assert.AreEqual("(BINARY.PEEK[1.1.TEXT]<1000.2000>)", ImapFetchDataItem.BinaryPeek("1.1.TEXT", new ImapPartialRange(1000L, 2000L)).ToString());
    }

    [Test]
    public void TestBinary_InvalidArgument()
    {
      Assert.Throws<ArgumentException>(() => ImapFetchDataItem.BinaryPeek("1.1", new ImapPartialRange(1000L)), "#1");
      Assert.Throws<ArgumentException>(() => ImapFetchDataItem.Binary    ("1.1", new ImapPartialRange(1000L)), "#2");
    }

    [Test]
    public void TestBinary_ArgumentOutOfRange()
    {
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.BinaryPeek("1.1", 1000L, 0L), "#1");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.Binary    ("1.1", 1000L, 0L), "#2");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.BinaryPeek("1.1", -1L, 1L), "#3");
      Assert.Throws<ArgumentOutOfRangeException>(() => ImapFetchDataItem.Binary    ("1.1", -1L, 1L), "#4");
    }

    [Test]
    public void TestBinary_SectionNull()
    {
      Assert.Throws<ArgumentNullException>(() => ImapFetchDataItem.Binary(null));
    }

    [Test]
    public void TestBinary_SectionEmpty()
    {
      Assert.Throws<ArgumentException>(() => ImapFetchDataItem.Binary(string.Empty));
    }

    [Test]
    public void TestCombine_ContainsRequiredCapability_1()
    {
      var dataItem1 = ImapFetchDataItem.ModSeq;
      var dataItem2 = ImapFetchDataItem.InternalDate;

      var combined = dataItem1 + dataItem2;

      Assert.AreEqual("(MODSEQ INTERNALDATE)", combined.ToString());
      CollectionAssert.AreEquivalent(new[] {ImapCapability.CondStore},
                                     (combined as IImapExtension).RequiredCapabilities);

      combined = dataItem2 + dataItem1;

      Assert.AreEqual("(INTERNALDATE MODSEQ)", combined.ToString());
      CollectionAssert.AreEquivalent(new[] {ImapCapability.CondStore},
                                     (combined as IImapExtension).RequiredCapabilities);
    }

    [Test]
    public void TestCombine_ContainsRequiredCapability_2()
    {
      var dataItem1 = ImapFetchDataItem.ModSeq;
      var dataItem2 = ImapFetchDataItem.Binary("1");

      var combined = dataItem1 + dataItem2;

      Assert.AreEqual("(MODSEQ BINARY[1])", combined.ToString());
      CollectionAssert.AreEquivalent(new[] {ImapCapability.CondStore, ImapCapability.Binary},
                                     (combined as IImapExtension).RequiredCapabilities);

      combined = dataItem2 + dataItem1;

      Assert.AreEqual("(BINARY[1] MODSEQ)", combined.ToString());
      CollectionAssert.AreEquivalent(new[] {ImapCapability.CondStore, ImapCapability.Binary},
                                     (combined as IImapExtension).RequiredCapabilities);
    }

    [Test]
    public void TestCombine_WithMacro_ContainsRequiredCapability()
    {
      var dataItem1 = ImapFetchDataItem.ModSeq;
      var dataItem2 = ImapFetchDataItem.Fast;

      var combined = dataItem1 + dataItem2;

      Assert.AreEqual("(MODSEQ FLAGS INTERNALDATE RFC822.SIZE)", combined.ToString());
      CollectionAssert.AreEquivalent(new[] {ImapCapability.CondStore},
                                     (combined as IImapExtension).RequiredCapabilities);

      combined = dataItem2 + dataItem1;

      Assert.AreEqual("(FLAGS INTERNALDATE RFC822.SIZE MODSEQ)", combined.ToString());
      CollectionAssert.AreEquivalent(new[] {ImapCapability.CondStore},
                                     (combined as IImapExtension).RequiredCapabilities);
    }
  }
}